"""Data info tools for FINAM."""

import copy
import datetime

import numpy as np

from ...errors import FinamMetaDataError
from ..grid_base import GridBase
from .mask import MASK_INDICATORS, Mask, mask_specified, masks_compatible, masks_equal
from .units import UNITS, compatible_units


def _format_mask(mask):
    if mask_specified(mask) and mask is not np.ma.nomask:
        return "<ndarray>"
    if mask is np.ma.nomask:
        return "nomask"
    return str(mask)


class Info:
    """Data info containing grid specification and metadata

    Parameters
    ----------
    time : datetime or None, optional
        time specification, default: None
    grid : Grid or NoGrid or None, optional
        grid specification, default: None
    meta : dict, optional
        dictionary of metadata, default: None
    mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray`, optional
        masking specification of the data. Options:
            * :any:`Mask.FLEX`: data can be masked or unmasked (default)
            * :any:`Mask.NONE`: data is unmasked and given as plain numpy array
            * valid boolean mask for MaskedArray
    **meta_kwargs
        additional metadata by name, will overwrite entries in ``meta``

    Attributes
    ----------
    grid : Grid or NoGrid or None
        grid specification
    meta : dict
        dictionary of metadata
    """

    def __init__(self, time=None, grid=None, meta=None, mask=Mask.FLEX, **meta_kwargs):
        self._time = self._grid = self._mask = None
        self.time = time
        self.grid = grid
        self.mask = mask
        # set meta last (see __setattr__)
        self.meta = meta or {}
        self.meta.update(meta_kwargs)
        # handle units
        units = self.meta.get("units", "")
        units = None if units is None else UNITS.Unit(units)
        self.meta["units"] = units

    @property
    def time(self):
        """datetime: current time."""
        return self._time

    @time.setter
    def time(self, time):
        if time is not None and not isinstance(time, datetime.datetime):
            msg = "Time in Info must be either None or a datetime"
            raise FinamMetaDataError(msg)
        self._time = time

    @property
    def grid(self):
        """Grid: data grid."""
        return self._grid

    @grid.setter
    def grid(self, grid):
        if grid is not None and not isinstance(grid, GridBase):
            msg = "Grid in Info must be either None or of a sub-class of GridBase"
            raise FinamMetaDataError(msg)
        self._grid = grid

    @property
    def mask(self):
        """Mask or ndarray: data mask."""
        return self._mask

    @mask.setter
    def mask(self, mask):
        if mask_specified(mask) and mask is not None:
            mask = np.ma.make_mask(mask, shrink=False)
            if (
                self.grid is not None
                and mask is not np.ma.nomask
                and not np.array_equal(self.grid_shape, np.shape(mask))
            ):
                msg = "Mask in Info not compatible with given grid."
                raise FinamMetaDataError(msg)
        self._mask = mask

    @property
    def grid_shape(self):
        """tuple: shape of the data grid."""
        return None if self.grid is None else self.grid.data_shape

    @property
    def is_masked(self):
        """bool: whether data is set to be masked."""
        return mask_specified(self.mask)

    @property
    def fill_value(self):
        """Fill value for masked data."""
        return self.meta.get(
            MASK_INDICATORS[0], self.meta.get(MASK_INDICATORS[1], None)
        )

    def copy(self):
        """Copies the info object"""
        return copy.copy(self)

    def copy_with(self, use_none=True, **kwargs):
        """Copies the info object and sets variables and meta values according to the kwargs

        Parameters
        ----------
        use_none : bool
            whether properties with None value should also be transferred
        **kwargs
            key values pairs for properties to change
        """
        other = Info(
            time=self.time, grid=self.grid, meta=copy.copy(self.meta), mask=self.mask
        )
        for k, v in kwargs.items():
            if k == "time":
                if v is not None or use_none:
                    other.time = v
            elif k == "grid":
                if v is not None or use_none:
                    other.grid = v
            elif k == "mask":
                if v is not None or use_none:
                    other.mask = v
            elif k == "units":
                if v is not None or use_none:
                    other.meta[k] = v if v is None else UNITS.Unit(v)
            else:
                if v is not None or use_none:
                    other.meta[k] = v

        return other

    def accepts(self, incoming, fail_info, incoming_donwstream=False):
        """
        Tests whether this info can accept/is compatible with an incoming info.

        Tested attributes are: "grid", "mask" and "units"

        Parameters
        ----------
        incoming : Info
            Incoming/source info to check. This is the info from upstream.
        fail_info : dict
            Dictionary that will be filled with failed properties; name: (source, target).
        incoming_donwstream : bool, optional
            Whether the incoming info is from downstream data. Default: False

        Returns
        -------
        bool
            Whether the incoming info is accepted
        """
        if not isinstance(incoming, Info):
            fail_info["type"] = (incoming.__class__, self.__class__)
            return False

        success = True
        if self.grid is not None and not self.grid.compatible_with(incoming.grid):
            if not (incoming_donwstream and incoming.grid is None):
                fail_info["grid"] = (incoming.grid, self.grid)
                success = False

        if self.mask is not None and not masks_compatible(
            self.mask, incoming.mask, incoming_donwstream, self.grid, incoming.grid
        ):
            if not (incoming_donwstream and incoming.mask is None):
                fail_info["mask"] = (incoming.mask, self.mask)
                success = False

        u1_none = (u1 := self.units) is None
        u2_none = (u2 := incoming.units) is None
        if not u1_none and (u2_none or not compatible_units(u1, u2)):
            if not (incoming_donwstream and u2_none):
                fail_info["units"] = (u2, u1)
                success = False

        return success

    def __copy__(self):
        """Shallow copy of the info"""
        return Info(time=self.time, grid=self.grid, meta=self.meta, mask=self.mask)

    def __eq__(self, other):
        """Equality check for two infos

        Ignores time.
        """
        if not isinstance(other, Info):
            return False
        return (
            self.grid == other.grid
            and self.meta == other.meta
            and masks_equal(self.mask, other.mask, self.grid, other.grid)
        )

    def __getattr__(self, name):
        # only called if attribute is not present in class
        if "meta" in self.__dict__ and name in self.meta:
            return self.meta[name]
        raise AttributeError(f"'Info' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        # first check if attribute present or meta not yet present (e.g. grid)
        if name in self.__dir__() or "meta" not in self.__dict__:
            super().__setattr__(name, value)
        else:
            self.__dict__["meta"][name] = value

    def __repr__(self):
        grid = self.grid.name if self.grid is not None else "None"
        meta = ", " * bool(self.meta)
        meta += ", ".join(
            f"{k}=" + ("None" if v is None else f"'{v}'") for k, v in self.meta.items()
        )
        return f"Info(grid={grid}, mask={_format_mask(self.mask)}{meta})"

    def as_dict(self):
        """Returns a ``dict`` containing all metadata in this Info."""
        return {
            **self.meta,
            "mask": _format_mask(self.mask),
            "grid": f"{self.grid}",
            "units": f"{self.units:~}",
        }



import unittest
from datetime import datetime, timedelta

import numpy as np
from numpy.testing import assert_allclose

import finam as fm


class TestHistogram(unittest.TestCase):
    def test_histogram(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((15, 12)),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=time,
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, sink])

        (
            source.outputs["Noise"]
            >> fm.adapters.Histogram(lower=-1, upper=1, bins=20)
            >> sink.inputs["Input"]
        )

        composition.connect()

        data = sink.data["Input"]
        self.assertEqual(data.shape, (1, 20))
        self.assertEqual(data.sum(), 11 * 14)
        self.assertEqual(fm.data.get_units(data), fm.UNITS.dimensionless)

        composition.run(end_time=datetime(2000, 1, 10))


if __name__ == "__main__":
    unittest.main()



"""
Unit tests for the adapters.probe module.
"""

import unittest
from datetime import datetime

from finam import UNITS, Info, Input, NoGrid, Output
from finam.adapters.probe import CallbackProbe


class TestProbe(unittest.TestCase):
    def test_probe_adapter(self):
        data = None
        time = None

        def callback(d, t):
            nonlocal data
            nonlocal time
            data = d
            time = t

        out = Output("test_out")
        adapter = CallbackProbe(callback=callback)
        inp = Input("test_in")

        out >> adapter >> inp

        out.push_info(Info(datetime(2000, 1, 1), grid=NoGrid(), units="m"))
        inp.exchange_info(Info(None, grid=NoGrid(), units=None))
        out.push_data(100 * UNITS.Unit("m"), datetime(2000, 1, 1))
        inp.pull_data(datetime(2000, 1, 1))

        self.assertEqual(data, 100 * UNITS.Unit("m"))
        self.assertEqual(time, datetime(2000, 1, 1))


if __name__ == "__main__":
    unittest.main()



"""Core data tools for FINAM."""

import copy
import datetime

import numpy as np
import pandas as pd

from ...errors import FinamDataError
from .. import grid_spec
from ..grid_base import Grid
from .units import (
    UNITS,
    check_quantified,
    compatible_units,
    equivalent_units,
    get_units,
    is_quantified,
)

_BASE_DATETIME = datetime.datetime(1970, 1, 1)
_BASE_TIME = np.datetime64("1970-01-01T00:00:00")
_BASE_DELTA = np.timedelta64(1, "s")


def prepare(data, info, time_entries=1, force_copy=False, report_conversion=False):
    """
    Prepares data in FINAM's internal transmission format.

    Checks tha shape of the data.
    Checks or adds units and time dimension.

    Parameters
    ----------
    data : arraylike
        The input data.
    info : Info
        Info associated with the data.
    time_entries : int, optional
        Number of time slices in the data. Default 1.
    force_copy : bool, optional
        Forces the result to be a copy of the passed data. Default ``False``.

        If not used, the result is a view of the data if no units conversion needs to be done.
    report_conversion : bool, optional
        If true, returns a tuple with the second element indicating the unit conversion if it was required.

    Returns
    -------
    pint.Quantity or tuple(pint.Quantity, tuple(pint.Unit, pint.Unit) or None)
        The prepared data as a numpy array, wrapped into a :class:`pint.Quantity`.

        If ``report_conversion`` is ``True``, a tuple is returned with the second element
        indicating the unit conversion if it was required.

        The second element is ``None`` if no conversion was required,
        and a tuple of two :class:`pint.Unit` objects otherwise.

    Raises
    ------
    FinamDataError
        If the data doesn't match its info.
    """
    units_converted = None
    units = info.units
    if is_quantified(data):
        if not compatible_units(data.units, units):
            raise FinamDataError(
                f"Given data has incompatible units. "
                f"Got {data.units}, expected {units}."
            )
        if info.is_masked and not np.ma.isarray(data.magnitude):
            data = UNITS.Quantity(
                np.ma.array(
                    data=data.magnitude,
                    mask=info.mask,
                    shrink=False,
                    fill_value=info.fill_value,
                ),
                data.units,
            )
        if not equivalent_units(data.units, units):
            units_converted = data.units, units
            data = data.to(units)
        elif force_copy:
            data = data.copy()
    else:
        if info.is_masked and not np.ma.isarray(data):
            data = UNITS.Quantity(
                np.ma.array(
                    data=data,
                    mask=info.mask,
                    shrink=False,
                    fill_value=info.fill_value,
                    copy=force_copy,
                ),
                units,
            )
        # this covers masked arrays as well
        elif isinstance(data, np.ndarray):
            if force_copy:
                data = data.copy()
            data = UNITS.Quantity(data, units)
        else:
            if force_copy:
                data = copy.copy(data)
            data = UNITS.Quantity(np.asarray(data), units)

    data = _check_input_shape(data, info, time_entries)

    if report_conversion:
        return data, units_converted
    return data


def _check_input_shape(data, info, time_entries):
    # check correct data size
    if isinstance(info.grid, Grid):
        time_entries = (
            data.shape[0]
            if len(data.shape) == len(info.grid.data_shape) + 1
            else time_entries
        )
        data_size = data.size / time_entries
        if data_size != info.grid.data_size:
            raise FinamDataError(
                f"quantify: data size doesn't match grid size. "
                f"Got {data_size}, expected {info.grid.data_size}"
            )
        # check shape of non-flat arrays
        if len(data.shape) != 1:
            if data.shape[1:] != info.grid.data_shape:
                if data.shape == info.grid.data_shape:
                    data = np.expand_dims(data, 0)
                else:
                    raise FinamDataError(
                        f"quantify: data shape doesn't match grid shape. "
                        f"Got {data.shape}, expected {info.grid.data_shape}"
                    )
        else:
            # reshape arrays
            if time_entries <= 1:
                data = data.reshape(
                    [1] + list(info.grid.data_shape), order=info.grid.order
                )
            else:
                data = data.reshape(
                    [time_entries] + list(info.grid.data_shape), order=info.grid.order
                )
    elif isinstance(info.grid, grid_spec.NoGrid):
        data = _check_input_shape_no_grid(data, info, time_entries)
    return data


def _check_input_shape_no_grid(data, info, time_entries):
    if len(data.shape) != info.grid.dim + 1:
        if _no_grid_shape_valid(data.shape, info.grid):
            data = np.expand_dims(data, 0)
        else:
            raise FinamDataError(
                f"Data shape not valid. "
                f"Got {data.shape}, expected {info.grid.data_shape}"
            )
    else:
        if not _no_grid_shape_valid(data.shape[1:], info.grid):
            raise FinamDataError(
                f"Data shape not valid. "
                f"Got {data.shape[1:]}, expected {info.grid.data_shape}"
            )
        if data.shape[0] != time_entries:
            raise FinamDataError(
                f"Number of time entries in data doesn't match expected number. "
                f"Got {data.shape[0]}, expected {time_entries}"
            )
    return data


def _no_grid_shape_valid(data_shape, grid):
    if len(data_shape) != grid.dim:
        return False
    dshp = np.array(data_shape)
    gshp = np.array(grid.data_shape)
    fix_dims = gshp != -1
    return np.all(dshp[fix_dims] == gshp[fix_dims])


def has_time_axis(xdata, grid):
    """
    Check if the data array has a time axis.

    Parameters
    ----------
    xdata : numpy.ndarray
        The given data array.
    grid : GridBase
        The associated grid specification
    Returns
    -------
    bool
        Whether the data has a time axis.
    """
    grid_dim = None

    if isinstance(grid, Grid):
        grid_dim = len(grid.data_shape)
    elif isinstance(grid, grid_spec.NoGrid):
        grid_dim = grid.dim
    else:
        raise ValueError(
            f"Expected type Grid or NoGrid, got {grid.__class__.__name__}."
        )

    if xdata.ndim == grid_dim:
        return False

    if xdata.ndim == grid_dim + 1:
        return True

    raise FinamDataError("Data dimension must be grid dimension or grid dimension + 1.")


def to_datetime(date):
    """Converts a numpy datetime64 object to a python datetime object"""
    if np.isnan(date):
        return pd.NaT

    timestamp = (date - _BASE_TIME) / _BASE_DELTA

    if timestamp < 0:
        return _BASE_DATETIME + datetime.timedelta(seconds=timestamp)

    tz = datetime.timezone.utc
    return datetime.datetime.fromtimestamp(timestamp, tz).replace(tzinfo=None)


def strip_time(xdata, grid):
    """Returns a view of the data with the time dimension squeezed if there is only a single entry

    Parameters
    ----------
    xdata : arraylike
        Data to strip time dimension from
    grid : GridBase
        The associated grid specification

    Returns
    -------
    arraylike
        Stripped data

    Raises
    ------
    FinamDataError
        If the data has multiple time entries.
    """
    if has_time_axis(xdata, grid):
        if xdata.shape[0] > 1:
            raise FinamDataError(
                "Can't strip time of a data array with multiple time entries"
            )
        return xdata[0, ...]

    return xdata


def full_like(xdata, value):
    """
    Return a new data array with the same shape, type and units as a given object.

    Parameters
    ----------
    xdata : :class:`pint.Quantity` or :class:`numpy.ndarray`
        The reference object input.
    value : scalar
        Value to fill the new object with before returning it.

    Returns
    -------
    pint.Quantity or numpy.ndarray
        New object with the same shape and type as other,
        with the data filled with fill_value.
        Units will be taken from the input if present.
    """
    data = np.full_like(xdata, value)
    if is_quantified(xdata):
        return UNITS.Quantity(data, xdata.units)
    return data


def full(value, info):
    """
    Return a new data array with units according to the given info, filled with given value.

    Parameters
    ----------
    value : scalar
        Value to fill the new object with before returning it.
    info : Info
        Info associated with the data.

    Returns
    -------
    pint.Quantity
        The converted data.
    """
    shape = info.grid.data_shape if isinstance(info.grid, Grid) else tuple()
    return prepare(np.full([1] + list(shape), value), info)


def check(xdata, info):
    """
    Check if data matches given info.

    Parameters
    ----------
    xdata : numpy.ndarray
        The given data array.
    info : Info
        Info associated with the data.

    Raises
    ------
    FinamDataError
        If data doesn't match given info.
    """
    check_quantified(xdata, "check")

    if not has_time_axis(xdata, info.grid):
        raise FinamDataError("check: given data should have a time dimension.")

    _check_shape(xdata.shape[1:], info.grid)

    # check units
    if not compatible_units(info.units, xdata):
        raise FinamDataError(
            f"check: given data has incompatible units. "
            f"Got {get_units(xdata)}, expected {info.units}."
        )


def _check_shape(shape, grid):
    if isinstance(grid, Grid) and shape != grid.data_shape:
        raise FinamDataError(
            f"check: given data has wrong shape. "
            f"Got {shape}, expected {grid.data_shape}"
        )
    if isinstance(grid, grid_spec.NoGrid) and len(shape) != grid.dim:
        raise FinamDataError(
            f"check: given data has wrong number of dimensions. "
            f"Got {len(shape)}, expected {grid.dim}"
        )


def assert_type(cls, slot, obj, types):
    """Type assertion."""
    for t in types:
        if isinstance(obj, t):
            return
    raise TypeError(
        f"Unsupported data type for {slot} in "
        f"{cls.__class__.__name__}: {obj.__class__.__name__}. "
        f"Expected one of [{', '.join([tp.__name__ for tp in types])}]"
    )



"""Common ESRI ASCII grid routines."""

import warnings

import numpy as np

ESRI_TYPES = {
    "ncols": int,
    "nrows": int,
    "xllcorner": float,
    "yllcorner": float,
    "xllcenter": float,
    "yllcenter": float,
    "cellsize": float,
    "nodata_value": float,
}
"""types for ESRI ASCII grid header information."""

ESRI_REQ = {"ncols", "nrows", "xllcorner", "yllcorner", "cellsize"}
"""Required ESRI ASCII grid header information."""


def _is_number(string):
    try:
        float(string)
        return True
    except ValueError:
        return False


def _extract_header(file):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return np.genfromtxt(
            file, dtype=str, max_rows=6, usecols=(0, 1), invalid_raise=False
        )


def standardize_header(header):
    """
    Standardize an ASCII grid header dictionary.

    Parameters
    ----------
    header : :class:`dict`
        Raw header as dictionary.

    Returns
    -------
    :class:`dict`
        Standardized header as dictionary.

    Raises
    ------
    ValueError
        If the header is missing required information.
        See :any:`ESRI_REQ`
    """
    header = {n: ESRI_TYPES[n](v) for (n, v) in header.items() if n in ESRI_TYPES}
    # convert cell center to corner information
    if "xllcenter" in header:
        header["xllcorner"] = header["xllcenter"] - 0.5 * header.get("cellsize", 1)
        del header["xllcenter"]
    if "yllcenter" in header:
        header["yllcorner"] = header["yllcenter"] - 0.5 * header.get("cellsize", 1)
        del header["yllcenter"]
    # check required header items
    missing = ESRI_REQ - (set(header) & ESRI_REQ)
    if missing:
        msg = f"standardize_header: missing header information {missing}"
        raise ValueError(msg)
    return header


def read_header(file):
    """
    Read an ASCII grid header from file.

    Parameters
    ----------
    file : :class:`~os.PathLike`
        File containing the ASCII grid header.

    Returns
    -------
    :class:`dict`
        Standardized header as dictionary.

    Notes
    -----
    "xllcenter" and "yllcenter" will be converted to
    "xllcorner" and "yllcorner" resepectively.
    """
    header_lines = _extract_header(file)
    return standardize_header(dict(header_lines))


def read_grid(file, dtype=None):
    """
    Read an ASCII grid from file.

    Parameters
    ----------
    file : :class:`~os.PathLike`
        File containing the ASCII grid.
    dtype : str/type, optional
        Data type.
        Needs to be integer or float and compatible with np.dtype
        (i.e. "i4", "f4", "f8"), by default None

    Returns
    -------
    header : dict
        Header describing the grid.
    data : numpy.ndarray
        Data of the grid.

    Raises
    ------
    ValueError
        If data shape is not matching the given header.
    """
    header_lines = _extract_header(file)
    header = standardize_header(dict(header_lines))
    # last line could already be data if "nodata_value" is missing
    numeric_last = _is_number(header_lines[-1][0])
    header_size = len(header_lines) - int(numeric_last)
    data = np.loadtxt(file, dtype=dtype, skiprows=header_size, ndmin=2)
    nrows, ncols = header["nrows"], header["ncols"]
    if data.shape[0] != nrows or data.shape[1] != ncols:
        msg = (
            f"read_grid: data shape {data.shape} "
            f"not matching given header ({nrows=}, {ncols=})."
        )
        raise ValueError(msg)
    if "nodata_value" in header and np.issubdtype(data.dtype, np.integer):
        header["nodata_value"] = int(header["nodata_value"])
    return header, data



import datetime
import unittest
from datetime import datetime as dt

import numpy as np
import pint

import finam
import finam.errors


class TestDataTools(unittest.TestCase):
    def test_check(self):
        time = dt(2000, 1, 1)

        grid = finam.EsriGrid(2, 3)
        gri0 = finam.EsriGrid(2, 3, axes_names=["lat", "lon"])
        gri1 = finam.EsriGrid(3, 2)

        info = finam.Info(time, grid, units="m", foo="bar")
        inf0 = info.copy_with(units="meter")
        inf1 = info.copy_with(foo="spam")
        inf2 = info.copy_with(units="s")
        inf3 = info.copy_with(grid=gri0)
        inf4 = info.copy_with(grid=gri1)

        time = dt(year=2022, month=10, day=12)
        tim0 = dt(year=2021, month=10, day=12)

        data = np.arange(6).reshape(3, 2)
        dar0 = finam.data.prepare(data, info)
        dar1 = finam.data.prepare(data, info)

        # assert stuff
        self.assertIsInstance(finam.data.get_magnitude(dar0), np.ndarray)
        self.assertIsInstance(finam.data.strip_time(dar0, info.grid), pint.Quantity)
        self.assertIsInstance(
            finam.data.get_dimensionality(dar0), pint.util.UnitsContainer
        )

        # should work
        finam.data.prepare(dar0, info)
        finam.data.prepare(dar1, info)
        finam.data.check(dar0, info)
        finam.data.check(dar1, info)
        finam.data.to_units(dar0, "km")

        # wrong shape
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.prepare(1, info)

        # no DataArray
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.check(None, info)

        # not qunatified
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.check(dar0.magnitude, info)

        finam.data.check(dar1, inf0)

        # other units format should work
        finam.data.check(dar0, inf0)

        # wrong units
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.check(dar0, inf2)

        # wrong shape
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.check(dar0, inf4)
        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.check(dar1, inf4)

        # check full_like
        dar2 = finam.data.full_like(dar0, 0)
        finam.data.check(dar2, info)

        dar3 = finam.data.full(0, info)
        finam.data.check(dar3, info)

    def test_other_grids(self):
        time = dt(2000, 1, 1)

        gri0 = finam.NoGrid(dim=1)
        gri1 = finam.UnstructuredPoints(points=[[0, 0], [0, 2], [2, 2]])
        info = finam.Info(time, gri0, units="s")
        data = np.arange(3)
        dar0 = finam.data.prepare(data, info)
        dar1 = finam.data.prepare(data, info.copy_with(grid=gri1))

        self.assertEqual((1, 3), dar0.shape)
        self.assertEqual((1, 3), dar1.shape)

    def test_strip_time(self):
        time = dt(2000, 1, 1)
        grid = finam.NoGrid()

        xdata = finam.data.prepare(1.0, finam.Info(time, grid=grid))
        self.assertEqual(xdata.shape, (1,))
        stripped = finam.data.strip_time(xdata, grid)
        self.assertEqual(stripped.shape, ())

        xdata = finam.data.prepare(
            [1.0, 2.0, 3.0],
            finam.Info(time, grid=finam.NoGrid(dim=1)),
        )
        self.assertEqual(xdata.shape, (1, 3))
        stripped = finam.data.strip_time(xdata, finam.NoGrid(dim=1))
        self.assertEqual(stripped.shape, (3,))
        stripped2 = finam.data.strip_time(xdata, finam.NoGrid(dim=1))
        self.assertEqual(stripped2.shape, stripped.shape)

        arr1 = finam.data.prepare(
            1.0,
            finam.Info(time, grid=finam.NoGrid()),
        )
        arr2 = finam.data.prepare(
            1.0,
            finam.Info(time, grid=finam.NoGrid()),
        )
        data = np.concatenate([arr1, arr2], axis=0)
        with self.assertRaises(finam.errors.FinamDataError):
            stripped_ = finam.data.strip_time(data, finam.NoGrid())

    def test_prepare(self):
        time = dt(2000, 1, 1)

        data = finam.data.prepare(1.0, finam.Info(time, grid=finam.NoGrid()))
        self.assertEqual(np.asarray([1.0]) * finam.UNITS(""), data)

        data = finam.data.prepare(
            [[1.0, 1.0], [1.0, 1.0]], finam.Info(time, grid=finam.UniformGrid((3, 3)))
        )
        self.assertEqual((1, 2, 2), data.shape)

        data = finam.data.prepare(
            finam.UNITS.Quantity(1.0, "m"),
            finam.Info(time, grid=finam.NoGrid(), units="m"),
        )

        self.assertEqual((1,), data.shape)
        self.assertEqual(finam.UNITS.meter, data.units)

        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.prepare(
                np.asarray([1, 2]), finam.Info(time, grid=finam.NoGrid())
            )

        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.prepare(
                1.0 * finam.UNITS.meter, finam.Info(time, grid=finam.NoGrid())
            )

        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.prepare(
                1.0 * finam.UNITS.meter,
                finam.Info(time, grid=finam.NoGrid(), units="m^3"),
            )

    def test_prepare_copy(self):
        time = dt(2000, 1, 1)
        info_1 = finam.Info(time, grid=finam.NoGrid(1), units="m")
        info_2 = finam.Info(time, grid=finam.NoGrid(1), units="km")

        # using numpy arrays without units
        data = np.asarray([1, 2])
        xdata = finam.data.prepare(data, info_1, force_copy=True)
        data[0] = 0
        self.assertNotEqual(xdata[0, 0], data[0])

        # using numpy arrays with units
        data = np.asarray([1, 2]) * finam.UNITS("m")
        xdata = finam.data.prepare(data, info_1)
        data[0] = 0 * finam.UNITS("m")
        self.assertEqual(xdata[0, 0], data[0])

        data = np.asarray([1, 2]) * finam.UNITS("m")
        xdata = finam.data.prepare(data, info_1, force_copy=True)
        data[0] = 0 * finam.UNITS("m")
        self.assertNotEqual(xdata[0, 0], data[0])

        data = np.asarray([1, 2]) * finam.UNITS("m")
        xdata = finam.data.prepare(data, info_2)
        data[0] = 0 * finam.UNITS("m")
        self.assertNotEqual(finam.data.get_magnitude(xdata[0, 0]), 0.0)

        xdata = finam.data.prepare(np.asarray([1, 2]), info_1)
        xdata2 = finam.data.prepare(xdata, info_1)
        xdata[0, 0] = 0 * finam.UNITS("m")
        self.assertEqual(xdata2[0, 0], xdata[0, 0])

        xdata = finam.data.prepare(np.asarray([1, 2]), info_1)
        xdata2 = finam.data.prepare(xdata, info_1, force_copy=True)
        xdata[0, 0] = 0 * finam.UNITS("m")
        self.assertNotEqual(xdata2[0, 0], xdata[0, 0])

        xdata = finam.data.prepare(np.asarray([1, 2]), info_1)
        xdata2 = finam.data.prepare(xdata, info_2)
        xdata[0, 0] = 0 * finam.UNITS("m")
        self.assertNotEqual(finam.data.get_magnitude(xdata2[0, 0]), 0.0)

        data = [1.0]
        xdata2 = finam.data.prepare(data, info_1, force_copy=True)
        self.assertEqual(1.0 * finam.UNITS.meter, xdata2[0])

        xdata2[0, 0] = 0 * finam.UNITS("m")
        self.assertNotEqual(0.0, data[0])

    def test_prepare_masked(self):
        time = dt(2000, 1, 1)

        info = finam.Info(
            time,
            grid=finam.UniformGrid((3, 4), data_location=finam.Location.POINTS),
            units="",
        )

        in_data = np.ma.MaskedArray(np.ndarray((3, 4)), mask=False)
        in_data.mask[0, 0] = True

        xdata = finam.data.prepare(in_data, info, force_copy=True)
        self.assertTrue(finam.data.is_masked_array(xdata))
        self.assertTrue(xdata.mask[0, 0, 0])
        self.assertFalse(xdata.mask[0, 1, 0])

        in_data = finam.data.quantify(in_data)
        xdata = finam.data.prepare(in_data, info, force_copy=True)
        self.assertTrue(finam.data.is_masked_array(xdata))
        self.assertTrue(xdata.mask[0, 0, 0])
        self.assertFalse(xdata.mask[0, 1, 0])

    def test_assert_type(self):
        finam.data.assert_type(self, "A", 1, [int, float])

        with self.assertRaises(TypeError):
            finam.data.assert_type(self, "A", "1", [int, float])

    def test_info(self):
        time = dt(2000, 1, 1)
        _info = finam.Info(time, grid=finam.NoGrid())

        with self.assertRaises(finam.errors.FinamMetaDataError):
            _info = finam.Info(time=1, grid=finam.NoGrid())

        with self.assertRaises(finam.errors.FinamMetaDataError):
            _info = finam.Info(time, grid=finam.NoGrid)

        info = finam.Info(time, grid=finam.NoGrid(), units="m", foo="bar")
        self.assertEqual(info, info.copy_with())

        self.assertFalse(info.accepts(0, {}))
        self.assertFalse(info.accepts(finam.Info(time=None, grid=None), {}))

        self.assertTrue(
            info.accepts(finam.Info(time, grid=finam.NoGrid(), units="m"), {})
        )
        self.assertTrue(
            info.accepts(finam.Info(time, grid=finam.NoGrid(), units="km"), {})
        )
        self.assertFalse(
            info.accepts(finam.Info(time, grid=finam.NoGrid(), units="s"), {})
        )

        self.assertTrue(
            info.accepts(
                finam.Info(time, grid=finam.NoGrid(), units="m", foo="baz"), {}
            )
        )

        self.assertEqual(info, info.copy())
        self.assertNotEqual(info, 0)

        with self.assertRaises(AttributeError):
            info.__getattr__("bar")

        info.__setattr__("bar", "baz")
        self.assertEqual(info.bar, "baz")

    def test_check_shape(self):
        time = dt(2000, 1, 1)
        xdata = finam.data.prepare(
            1.0,
            finam.Info(time, grid=finam.NoGrid()),
        )

        finam.data.tools.core._check_shape(xdata.shape[1:], finam.NoGrid())

        with self.assertRaises(finam.errors.FinamDataError):
            finam.data.tools.core._check_shape(xdata.shape[1:], finam.NoGrid(dim=1))

    def test_quantify(self):
        xdata = np.asarray([1.0])
        xdata = finam.data.quantify(xdata, "m")
        self.assertEqual(finam.data.get_units(xdata), finam.UNITS.meter)

        xdata = np.asarray([1.0])
        xdata = finam.data.quantify(xdata)
        self.assertEqual(finam.data.get_units(xdata), finam.UNITS.dimensionless)

        xdata = np.asarray([1.0]) * finam.UNITS("")
        with self.assertRaises(finam.FinamDataError):
            xdata = finam.data.quantify(xdata)

    def test_to_datetime(self):
        t = np.datetime64("1900-01-01")
        self.assertEqual(datetime.datetime(1900, 1, 1), finam.data.to_datetime(t))

        t = np.datetime64("2000-01-01")
        self.assertEqual(datetime.datetime(2000, 1, 1), finam.data.to_datetime(t))

    def test_cache_units(self):
        finam.data.tools.clear_units_cache()

        self.assertEqual({}, finam.data.tools.units._UNIT_PAIRS_CACHE)

        eqiv = finam.data.tools.equivalent_units("mm", "L/m^2")
        self.assertTrue(eqiv)
        self.assertEqual(
            {(finam.UNITS.Unit("mm"), finam.UNITS.Unit("L/m^2")): (True, True)},
            finam.data.tools.units._UNIT_PAIRS_CACHE,
        )

        finam.data.tools.clear_units_cache()

        self.assertEqual({}, finam.data.tools.units._UNIT_PAIRS_CACHE)


if __name__ == "__main__":
    unittest.main()



"""
Adapters to transform or interpolate data when passed between components.

See also book chapter :doc:`/finam-book/usage/known_modules` for a list of
other adapters that are not included in the core package.

Base adapters
=============

.. autosummary::
   :toctree: generated

    Callback
    Scale
    ValueToGrid
    GridToValue

Probe adapters
==============

.. autosummary::
   :toctree: generated

    CallbackProbe

Regridding adapters
===================

.. autosummary::
   :toctree: generated

    RegridNearest
    RegridLinear

Statistics adapters
===================

.. autosummary::
   :toctree: generated

    Histogram

Time adapters
=============

.. autosummary::
   :toctree: generated

    LinearTime
    NextTime
    PreviousTime
    StackTime
    StepTime
    AvgOverTime
    SumOverTime
    DelayFixed
    DelayToPush
    DelayToPull
    TimeCachingAdapter
"""

from . import base, probe, regrid, time
from .base import Callback, GridToValue, Scale, ValueToGrid
from .probe import CallbackProbe
from .regrid import RegridLinear, RegridNearest
from .stats import Histogram
from .time import (
    DelayFixed,
    DelayToPull,
    DelayToPush,
    LinearTime,
    NextTime,
    PreviousTime,
    StackTime,
    StepTime,
    TimeCachingAdapter,
)
from .time_integration import AvgOverTime, SumOverTime

__all__ = ["base", "probe", "regrid", "stats", "time"]
__all__ += [
    "Callback",
    "Scale",
    "ValueToGrid",
    "GridToValue",
]
__all__ += ["CallbackProbe"]
__all__ += [
    "RegridNearest",
    "RegridLinear",
]
__all__ += ["Histogram"]
__all__ += [
    "NextTime",
    "PreviousTime",
    "LinearTime",
    "IntegrateTime",
    "StackTime",
    "StepTime",
    "DelayFixed",
    "DelayToPush",
    "DelayToPull",
    "TimeCachingAdapter",
]
__all__ += [
    "AvgOverTime",
    "SumOverTime",
]



"""
Unit tests for the adapters.time module.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np
import pint
from numpy.testing import assert_array_equal

from finam import FinamTimeError, Info, NoGrid, UniformGrid
from finam import data as tools
from finam.adapters.time import (
    DelayFixed,
    DelayToPull,
    DelayToPush,
    LinearTime,
    NextTime,
    PreviousTime,
    StackTime,
    StepTime,
    interpolate,
    interpolate_step,
)
from finam.components import CallbackGenerator

reg = pint.UnitRegistry(force_ndarray_like=True)


class TestDelayToPush(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(days=1),
        )

        self.adapter = DelayToPush()

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_delay_to_push(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 5, 0), None), 0.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 5, 0), None), 1.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 5, 0), None), 2.0)


class TestDelayToPull(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(days=1),
        )

        self.adapter = DelayToPull(steps=2, additional_delay=timedelta(days=0.8))

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_delay_to_pull(self):
        for _ in range(10):
            self.source.update()

        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 1)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 0), None), 1)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3, 0), None), 1)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 4, 0), None), 1)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 8, 0), None), 2)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 12, 0), None), 3)


class TestDelayFixed(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.last_pull = None

        def callback(t):
            return t.day - 1

        self.source = CallbackGenerator(
            callbacks={"Step": (callback, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(days=1),
        )

        self.adapter = DelayFixed(delay=timedelta(days=10))

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_fixed_delay(self):
        data = self.adapter.get_data(datetime(2000, 1, 1), None)
        self.assertEqual(data, 0)

        self.source.update()
        self.source.update()

        data = self.adapter.get_data(datetime(2000, 1, 5), None)
        self.assertEqual(data, 0)

        for _ in range(20):
            self.source.update()

        data = self.adapter.get_data(datetime(2000, 1, 10), None)
        self.assertEqual(data, 0)

        data = self.adapter.get_data(datetime(2000, 1, 11), None)
        self.assertEqual(data, 0)

        data = self.adapter.get_data(datetime(2000, 1, 12), None)
        self.assertEqual(data, 1)

        data = self.adapter.get_data(datetime(2000, 1, 20), None)
        self.assertEqual(data, 9)


class TestNextValue(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = NextTime()

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_next_value(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 12), None), 1.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 0), None), 1.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 12), None), 2.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3, 0), None), 2.0)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 4, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestPreviousValue(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(days=1),
        )

        self.adapter = PreviousTime()

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter
        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_previous_value(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 12), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 0), None), 1.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 12), None), 1.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3, 0), None), 2.0)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 1, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 4, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestInterpolation(unittest.TestCase):
    def test_linear(self):
        self.assertEqual(interpolate(0.0, 10.0, 0.0), 0.0)
        self.assertEqual(interpolate(0.0, 10.0, 1.0), 10.0)
        self.assertEqual(interpolate(0.0, 10.0, 0.1), 1.0)

    def test_step(self):
        self.assertEqual(interpolate_step(0.0, 10.0, 0.0, 0.3), 0.0)
        self.assertEqual(interpolate_step(0.0, 10.0, 1.0, 0.3), 10.0)
        self.assertEqual(interpolate_step(0.0, 10.0, 0.1, 0.3), 0.0)
        self.assertEqual(interpolate_step(0.0, 10.0, 0.3, 0.3), 0.0)
        self.assertEqual(interpolate_step(0.0, 10.0, 0.31, 0.3), 10.0)
        self.assertEqual(interpolate_step(0.0, 10.0, 1.0, 0.3), 10.0)


class TestStepInterpolation(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(days=10.0),
        )

        self.adapter = StepTime(step=0.3)

        self.source.initialize()
        self.source.outputs["Step"] >> self.adapter
        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_step_interpolation(self):
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 4, 0), None), 0.0)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 5, 0), None), 10.0)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 12, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestLinearInterpolation(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = LinearTime()

        self.source.initialize()
        self.source.outputs["Step"] >> self.adapter
        self.adapter.get_info(Info(None, grid=NoGrid()))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_linear_interpolation(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 0), None), 0.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1, 12), None), 0.5)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 0), None), 1.0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2, 12), None), 1.5)
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3, 0), None), 2.0)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 1, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 4, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestLinearGridInterpolation(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        grid, _ = create_grid(10, 15, 0)
        self.source = CallbackGenerator(
            callbacks={
                "Grid": (
                    lambda t: create_grid(10, 15, t.day - 1)[1],
                    Info(None, grid=grid),
                )
            },
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = LinearTime()

        self.source.initialize()
        self.source.outputs["Grid"] >> self.adapter
        self.adapter.get_info(Info(None, grid=grid))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_linear_grid_interpolation(self):
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 1, 0), None))[
                0, 2, 3
            ],
            0.0,
        )
        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 1, 12), None))[
                0, 2, 3
            ],
            0.5,
        )
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 2, 0), None))[
                0, 2, 3
            ],
            1.0,
        )
        self.source.update()
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 2, 12), None))[
                0, 2, 3
            ],
            1.5,
        )
        self.assertEqual(
            tools.get_magnitude(self.adapter.get_data(datetime(2000, 1, 3, 0), None))[
                0, 2, 3
            ],
            2.0,
        )

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 1, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(datetime(2000, 1, 4, 0), None)

        with self.assertRaises(FinamTimeError):
            self.adapter.get_data(100, None)


class TestTimeStack(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        grid, _ = create_grid(10, 15, 0)
        self.source = CallbackGenerator(
            callbacks={
                "Grid": (
                    lambda t: create_grid(10, 15, t.day - 1)[1],
                    Info(None, grid=grid),
                )
            },
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = StackTime()

        self.source.initialize()

        self.source.outputs["Grid"] >> self.adapter
        self.adapter.get_info(Info(None, grid=grid))

        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_stack_time(self):
        self.source.update()
        self.source.update()
        self.source.update()

        data = self.adapter.get_data(datetime(2000, 1, 4), None)
        self.assertEqual(data.shape, (4, 10, 15))

        self.source.update()
        self.source.update()

        data = self.adapter.get_data(datetime(2000, 1, 6), None)
        self.assertEqual(data.shape, (3, 10, 15))


def create_grid(cols, rows, value):
    grid = UniformGrid((cols, rows), data_location="POINTS")

    data = np.full(shape=grid.data_shape, fill_value=value, order=grid.order)

    return grid, data


if __name__ == "__main__":
    unittest.main()



"""
Implementations of IOutput
"""

import logging
import os
from datetime import datetime

import numpy as np

from ..data import tools
from ..data.tools import Info
from ..errors import (
    FinamDataError,
    FinamMetaDataError,
    FinamNoDataError,
    FinamStaticDataError,
    FinamTimeError,
)
from ..interfaces import IAdapter, IInput, IOutput, Loggable
from ..tools.log_helper import ErrorLogger


# pylint: disable=too-many-public-methods
class Output(IOutput, Loggable):
    """Default output implementation."""

    def __init__(self, name=None, info=None, static=False, **info_kwargs):
        Loggable.__init__(self)
        self._targets = []
        self.data = []
        self._output_info = None
        self.base_logger_name = None
        if name is None:
            raise ValueError("Output: needs a name.")
        self._name = name
        self._static = static

        if info_kwargs:
            if info is not None:
                raise ValueError("Output: can't use **kwargs in combination with info")
            info = Info(**info_kwargs)
        if info is not None:
            self.push_info(info)

        self._connected_inputs = {}
        self._out_infos_exchanged = 0

        self._time = None
        self._mem_limit = None
        self._mem_location = None
        self._total_mem = 0
        self._mem_counter = 0

    @property
    def name(self):
        """Input name."""
        return self._name

    @property
    def time(self):
        """The output's time of the latest available data"""
        return self._time

    @property
    def is_static(self):
        """Whether the output is static"""
        return self._static

    @property
    def info(self):
        """Info: The input's data info."""
        if self._output_info is None:
            raise FinamNoDataError("No data info available")
        if self.has_targets and self._out_infos_exchanged < len(self._connected_inputs):
            raise FinamNoDataError("Data info was not completely exchanged yet")

        return self._output_info

    @property
    def needs_pull(self):
        """bool: if the output needs pull."""
        return False

    @property
    def needs_push(self):
        """bool: if the output needs push."""
        return True

    @property
    def memory_limit(self):
        """The memory limit for this slot"""
        return self._mem_limit

    @memory_limit.setter
    def memory_limit(self, limit):
        """The memory limit for this slot"""
        self._mem_limit = limit

    @property
    def memory_location(self):
        """The memory-mapping location for this slot"""
        return self._mem_location

    @memory_location.setter
    def memory_location(self, directory):
        """The memory-mapping location for this slot"""
        self._mem_location = directory

    def has_info(self):
        """Returns if the output has a data info.

        The info is not required to be validly exchanged.
        """
        return self._output_info is not None

    def add_target(self, target):
        """Add a target input or adapter for this output.

        Parameters
        ----------
        target : :class:`.IInput`
            The target to add.
        """
        self.logger.trace("add target")
        if not isinstance(target, IInput):
            with ErrorLogger(self.logger):
                raise ValueError("Only IInput can added as target for IOutput")

        self._targets.append(target)

    @property
    def targets(self):
        """Get target inputs and adapters for this output.

        Returns
        -------
        list
            List of targets.
        """
        return self._targets

    def pinged(self, source):
        """Called when receiving a ping from a downstream input."""
        if not isinstance(source, IAdapter) and source in self._connected_inputs:
            with ErrorLogger(self.logger):
                raise ValueError(
                    f"Input '{source.name}' is already connected to this output"
                )
        self._connected_inputs[source] = None

    def push_data(self, data, time):
        """Push data into the output.

        Should notify targets, and can handle the provided date.

        Parameters
        ----------
        data : array_like
            Data set to push.
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the data set.
        """
        if not self.has_targets:
            self.logger.trace("skipping push to unconnected output")
            return

        self.logger.trace("push data")

        with ErrorLogger(self.logger):
            _check_time(time, self.is_static)

        if self.has_targets and self._out_infos_exchanged < len(self._connected_inputs):
            raise FinamNoDataError("Can't push data before output info was exchanged.")

        if self.is_static:
            if len(self.data) > 0:
                raise FinamStaticDataError(
                    "Can't push data repeatedly to a static output."
                )
            time = None

        with ErrorLogger(self.logger):
            xdata, conv = tools.prepare(data, self.info, report_conversion=True)
            if len(self.data) > 0 and not isinstance(self.data[-1][1], str):
                d = self.data[-1][1]
                if np.may_share_memory(d.data, xdata.data):
                    raise FinamDataError(
                        "Received data that shares memory with previously received data."
                    )
            if conv is not None:
                self.logger.profile(
                    "converted units from %s to %s (%d entries)", *conv, xdata.size
                )
            xdata = self._pack(xdata)
            self.data.append((time, xdata))

        self._time = time

        self.logger.trace("data cache: %d", len(self.data))

        self.notify_targets(time)

    def push_info(self, info):
        """Push data info into the output.

        Parameters
        ----------
        info : :class:`.Info`
            Delivered data info
        """
        self.logger.trace("push info")
        if not isinstance(info, Info):
            with ErrorLogger(self.logger):
                raise FinamMetaDataError("Metadata must be of type Info")
        self._output_info = info

    def notify_targets(self, time):
        """Notify all targets by calling their ``source_updated(time)`` method.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the simulation.
        """
        self.logger.trace("notify targets")

        with ErrorLogger(self.logger):
            _check_time(time, self.is_static)

        for target in self._targets:
            target.source_updated(time)

    def get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            simulation time to get the data for.
        target : :class:`.IInput` or None
            Requesting end point of this pull.

        Returns
        -------
        :class:`pint.Quantity`
            data-set for the requested time.

        Raises
        ------
        FinamNoDataError
            Raises the error if no data is available
        """
        self.logger.trace("get data")

        with ErrorLogger(self.logger):
            _check_time(time, self.is_static)

        if self._output_info is None:
            raise FinamNoDataError(f"No data info available in {self.name}")
        if self._out_infos_exchanged < len(self._connected_inputs):
            raise FinamNoDataError(f"Data info was not yet exchanged in {self.name}")
        if len(self.data) == 0:
            raise FinamNoDataError(f"No data available in {self.name}")

        with ErrorLogger(self.logger):
            data = (
                self._unpack(self.data[0][1])
                if self.is_static
                else self._interpolate(time)
            )

        if not self.is_static:
            data_count = len(self.data)
            self._clear_data(time, target)

            if len(self.data) < data_count:
                self.logger.trace(
                    "reduced data cache: %d -> %d", data_count, len(self.data)
                )

        return data

    def _pack(self, data):
        data_size = data.nbytes
        if self.memory_limit is not None and 0 <= self.memory_limit < (
            self._total_mem + data_size
        ):
            fn = os.path.join(
                self.memory_location or "", f"{id(self)}-{self._mem_counter}.npy"
            )
            self.logger.profile(
                "dumping data to file %s (total RAM %0.2f MB)",
                fn,
                self._total_mem / 1048576,
            )
            self._mem_counter += 1
            np.save(fn, data.magnitude)
            return fn

        self._total_mem += data_size
        self.logger.trace(
            "keeping data in RAM (total RAM %0.2f MB)", self._total_mem / 1048576
        )
        return data

    def _unpack(self, where):
        if isinstance(where, str):
            self.logger.profile("reading data from file %s", where)
            data = np.load(where, allow_pickle=True)
            return tools.UNITS.Quantity(data, self.info.units)

        return where

    def _clear_data(self, time, target):
        self._connected_inputs[target] = time
        if any(t is None for t in self._connected_inputs.values()):
            return

        t_min = min(self._connected_inputs.values())
        while len(self.data) > 1 and self.data[1][0] <= t_min:
            d = self.data.pop(0)
            if isinstance(d[1], str):
                os.remove(d[1])
            else:
                self._total_mem -= d[1].nbytes

    def finalize(self):
        """Finalize the output"""
        for _t, d in self.data:
            if isinstance(d, str):
                os.remove(d)
        self.data.clear()

    def _interpolate(self, time):
        if time < self.data[0][0] or time > self.data[-1][0]:
            raise FinamTimeError(
                f"Requested time {time} out of range [{self.data[0][0]}, {self.data[-1][0]}]"
            )
        for i, (t, data) in enumerate(self.data):
            if time > t:
                continue
            if time == t:
                return self._unpack(data)

            t_prev, data_prev = self.data[i - 1]
            diff = t - t_prev
            t_half = t_prev + diff / 2

            if time < t_half:
                return self._unpack(data_prev)

            return self._unpack(data)

        raise FinamTimeError(
            f"Time interpolation failed. This should not happen and is probably a bug. "
            f"Time is {time}."
        )

    def get_info(self, info):
        """Exchange and get the output's data info.

        For internal use. To get the info in a component, use property :attr:`.info`.

        Parameters
        ----------
        info : :class:`Info`
            Requested data info

        Returns
        -------
        dict
            Delivered data info

        Raises
        ------
        FinamNoDataError
            Raises the error if no info is available
        """
        self.logger.trace("get info")

        if self._output_info is None:
            raise FinamNoDataError("No data info available")

        fail_info = {}
        if not self._output_info.accepts(info, fail_info, incoming_donwstream=True):
            fail_info = "\n".join(
                [
                    f"{name} - got {got}, expected {exp}"
                    for name, (got, exp) in fail_info.items()
                ]
            )
            with ErrorLogger(self.logger):
                raise FinamMetaDataError(
                    f"Can't accept conflicting data infos. Failed entries:\n{fail_info}"
                )

        with ErrorLogger(self.logger):
            if self._output_info.grid is None:
                if info.grid is None:
                    raise FinamMetaDataError(
                        "Can't set property `grid` from target info, as it is not provided"
                    )

                self._output_info.grid = info.grid

            if self._output_info.time is None:
                if not self.is_static and info.time is None:
                    raise FinamMetaDataError(
                        "Can't set property `time` from target info, as it is not provided"
                    )

                self._output_info.time = info.time

            for k, v in self._output_info.meta.items():
                if v is None:
                    if k not in info.meta or info.meta[k] is None:
                        raise FinamMetaDataError(
                            f"Can't set property `meta.{k}` from target info, as it is not provided"
                        )

                    self._output_info.meta[k] = info.meta[k]

        self._out_infos_exchanged += 1

        return self._output_info

    def chain(self, other):
        """Chain outputs and adapters.

        Parameters
        ----------
        other : :class:`.IInput` or :class:`.IAdapter`
            The adapter or input to add as target to this output.

        Returns
        -------
        :class:`.IOutput`
            The last element of the chain.
        """
        self.logger.trace("chain")
        self.add_target(other)
        other.source = self
        return other

    @property
    def has_targets(self):
        """Flag if this output instance has any targets."""
        return bool(self._targets)

    @property
    def logger_name(self):
        """Logger name derived from base logger name and class name."""
        base_logger = logging.getLogger(self.base_logger_name)
        # logger hierarchy indicated by "." in name
        return ".".join(([base_logger.name, "->", self.name]))

    @property
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute. True."""
        return True


class CallbackOutput(Output):
    """Output implementation calling a callback when pulled.

    Use for components without time step.

    Parameters
    ----------
    callback : callable
        A callback ``callback(data, time)``, returning the transformed data.
    """

    def __init__(self, callback, name, info=None, **info_kwargs):
        super().__init__(name=name, info=info, static=False, **info_kwargs)
        self.callback = callback
        self.last_data = None

    @property
    def needs_push(self):
        """bool: if the output needs push."""
        return False

    @property
    def needs_pull(self):
        """bool: if the output needs pull."""
        return True

    def push_data(self, data, time):
        raise NotImplementedError("CallbackInput does not support push of data")

    def get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput`
            Requesting end point of this pull

        Returns
        -------
        :class:`pint.Quantity`
            Data-set for the requested time.

        Raises
        ------
        FinamNoDataError
            Raises the error if no data is available
        """
        self.logger.trace("get data")

        with ErrorLogger(self.logger):
            _check_time(time, False)

        if self._output_info is None:
            raise FinamNoDataError(f"No data info available in {self.name}")
        if self._out_infos_exchanged < len(self._connected_inputs):
            raise FinamNoDataError(f"Data info was not yet exchanged in {self.name}")

        data = self.callback(self, time)

        if data is None:
            raise FinamNoDataError(f"No data available in {self.name}")

        with ErrorLogger(self.logger):
            xdata, conv = tools.prepare(data, self.info, report_conversion=True)
            if self.last_data is not None and np.may_share_memory(
                tools.get_magnitude(self.last_data), tools.get_magnitude(xdata)
            ):
                raise FinamDataError(
                    "Received data that shares memory with previously received data."
                )
            if conv is not None:
                self.logger.profile(
                    "converted units from %s to %s (%d entries)", *conv, xdata.size
                )
            self.last_data = xdata
            return xdata

    def finalize(self):
        """Finalize the output"""


def _check_time(time, is_static):
    if is_static:
        if time is not None and not isinstance(time, datetime):
            raise ValueError("Time must be of type datetime or None in static outputs")
    else:
        if not isinstance(time, datetime):
            raise ValueError("Time must be of type datetime")



"""
Unit tests for the adapters.base module.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np
from numpy.testing import assert_allclose

import finam as fm
from finam import UNITS, Info, NoGrid, UniformGrid
from finam import data as fmdata
from finam.adapters.base import Callback, GridToValue, Scale, ValueToGrid
from finam.components.generators import CallbackGenerator


class TestCallback(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)

        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = Callback(callback=lambda v, t: v * 2)

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))
        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_callback_adapter(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1), None), 0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2), None), 2)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3), None), 4)


class TestScale(unittest.TestCase):
    def setUp(self):
        start = datetime(2000, 1, 1)
        self.source = CallbackGenerator(
            callbacks={"Step": (lambda t: t.day - 1, Info(None, grid=NoGrid()))},
            start=start,
            step=timedelta(1.0),
        )

        self.adapter = Scale(scale=2.0)

        self.source.initialize()

        self.source.outputs["Step"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid()))
        self.source.connect(start)
        self.source.connect(start)
        self.source.validate()

    def test_callback_adapter(self):
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 1), None), 0)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 2), None), 2)
        self.source.update()
        self.assertEqual(self.adapter.get_data(datetime(2000, 1, 3), None), 4)


class TestGridToValue(unittest.TestCase):
    def setUp(self):
        grid, data = create_grid(20, 10, 1.0)
        mask = np.full_like(data, False, dtype=bool)
        mask[0, :] = True

        self.source = CallbackGenerator(
            callbacks={
                "Grid": (lambda t: data, Info(None, grid=grid, mask=mask, units="m"))
            },
            start=datetime(2000, 1, 1),
            step=timedelta(1.0),
        )

        self.source.initialize()

    def test_grid_to_value_mean(self):
        self.adapter = GridToValue(func=np.ma.mean)
        self.source.outputs["Grid"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid(), units=None))
        self.source.connect(datetime(2000, 1, 1))
        self.source.connect(datetime(2000, 1, 1))
        self.source.validate()

        result = self.adapter.get_data(datetime(2000, 1, 1), None)
        self.assertEqual(result, 1.0 * UNITS.meter)

    def test_grid_to_value_sum(self):
        self.adapter = GridToValue(func=np.ma.sum)
        self.source.outputs["Grid"] >> self.adapter

        self.adapter.get_info(Info(None, grid=NoGrid(), units=None))
        self.source.connect(datetime(2000, 1, 1))
        self.source.connect(datetime(2000, 1, 1))
        self.source.validate()

        result = self.adapter.get_data(datetime(2000, 1, 1), None)
        self.assertEqual(result, 190.0 * UNITS.meter)


class TestValueToGrid(unittest.TestCase):
    def setUp(self):
        self.source = CallbackGenerator(
            callbacks={
                "Value": (
                    lambda t: 1.0,
                    Info(None, grid=NoGrid(), units="m"),
                )
            },
            start=datetime(2000, 1, 1),
            step=timedelta(1.0),
        )

        self.source.initialize()

    def test_value_to_grid(self):
        grid, data = create_grid(10, 10, 1.0)

        self.adapter = ValueToGrid(grid)
        self.source.outputs["Value"] >> self.adapter

        self.adapter.get_info(Info(None, grid=grid, units=None))
        self.source.connect(datetime(2000, 1, 1))
        self.source.connect(datetime(2000, 1, 1))
        self.source.validate()

        _reference_grid, reference_data = create_grid(10, 10, 1.0)
        out_data = self.adapter.get_data(datetime(2000, 1, 1), None)

        assert_allclose(fmdata.get_magnitude(out_data)[0, ...], reference_data)
        self.assertEqual(fmdata.get_units(out_data), UNITS("m"))

    def test_value_to_masked_grid(self):
        grid, data = create_grid(10, 10, 1.0)
        mask = np.full_like(data, False, dtype=bool)
        mask[0, :] = True

        self.adapter = ValueToGrid(grid)
        self.source.outputs["Value"] >> self.adapter

        self.adapter.get_info(Info(None, grid=grid, mask=mask, units=None))
        self.source.connect(datetime(2000, 1, 1))
        self.source.connect(datetime(2000, 1, 1))
        self.source.validate()

        _reference_grid, reference_data = create_grid(10, 10, 1.0)
        out_data = self.adapter.get_data(datetime(2000, 1, 1), None)
        reference_data = np.ma.array(reference_data, mask=mask)
        assert_allclose(fmdata.get_magnitude(out_data)[0, ...], reference_data)
        self.assertEqual(fmdata.get_units(out_data), UNITS("m"))

    def test_value_to_grid_wrong_grid(self):
        grid, data = create_grid(10, 10, 1.0)

        self.adapter = ValueToGrid(grid)
        self.source.outputs["Value"] >> self.adapter

        with self.assertRaises(fm.FinamMetaDataError):
            self.adapter.get_info(Info(None, grid=UniformGrid((2, 2)), units=None))

    def test_value_to_grid_from_target(self):
        grid, data = create_grid(10, 10, 1.0)

        self.adapter = ValueToGrid(None)
        self.source.outputs["Value"] >> self.adapter

        self.adapter.get_info(Info(None, grid=grid, units=None))
        self.source.connect(datetime(2000, 1, 1))
        self.source.connect(datetime(2000, 1, 1))
        self.source.validate()

        _reference_grid, reference_data = create_grid(10, 10, 1.0)
        out_data = self.adapter.get_data(datetime(2000, 1, 1), None)

        assert_allclose(fmdata.get_magnitude(out_data)[0, ...], reference_data)
        self.assertEqual(fmdata.get_units(out_data), UNITS("m"))


def create_grid(cols, rows, value):
    grid = UniformGrid((cols, rows), data_location="POINTS")

    data = np.full(shape=grid.data_shape, fill_value=value, order=grid.order)

    return grid, data


if __name__ == "__main__":
    unittest.main()



"""
Abstract base implementations for components with and without time step.
"""
import collections
import logging
from abc import ABC
from datetime import datetime
from enum import IntEnum
from typing import final

from ..errors import FinamLogError, FinamStatusError, FinamTimeError
from ..interfaces import (
    ComponentStatus,
    IComponent,
    IInput,
    IOutput,
    ITimeComponent,
    Loggable,
)
from ..tools.connect_helper import ConnectHelper
from ..tools.enum_helper import get_enum_value
from ..tools.log_helper import ErrorLogger, is_loggable
from .input import Input
from .output import Output


class Component(IComponent, Loggable, ABC):
    """Abstract component implementation.

    Extend this class for components without time step.
    See :doc:`/finam-book/development/special_components`.
    For components with a time step, use :class:`.TimeComponent`.

    Derived classes overwrite these methods:

    * :meth:`._initialize`
    * :meth:`._connect`
    * :meth:`._validate`
    * :meth:`._update`
    * :meth:`._finalize`
    """

    def __init__(self):
        Loggable.__init__(self)
        self._name = self.__class__.__name__
        self._status = ComponentStatus.CREATED
        self._inputs = IOList(self, "INPUT")
        self._outputs = IOList(self, "OUTPUT")
        self.base_logger_name = None
        self._connector: ConnectHelper = None

    def with_name(self, name):
        """Renames the component and returns self."""
        self._name = name
        return self

    @final
    def initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have :attr:`.status` :attr:`.ComponentStatus.INITIALIZED`.
        """
        self.logger.debug("init")
        self._initialize()

        self.inputs.frozen = True
        self.outputs.frozen = True

        if self.status != ComponentStatus.FAILED:
            self.status = ComponentStatus.INITIALIZED

    def _initialize(self):
        """Initialize the component.

        Components must overwrite this method.
        After the method call, the component's inputs and outputs must be available.
        """
        raise NotImplementedError(
            f"Method `_initialize` must be implemented by all components, but implementation is missing in {self.name}."
        )

    @final
    def connect(self, start_time):
        """Connect exchange data and metadata with linked components.

        The method can be called multiple times if there are failed pull attempts.

        After each method call, the component should have :attr:`.status` :attr:`.ComponentStatus.CONNECTED` if
        connecting was completed, :attr:`.ComponentStatus.CONNECTING` if some but not all required initial input(s)
        could be pulled, and :attr:`.ComponentStatus.CONNECTING_IDLE` if nothing could be pulled.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`
            The composition's starting time.
            Can be before the component's actual time.
        """
        if start_time is not None and not isinstance(start_time, datetime):
            raise FinamTimeError("Time in connect must be either None or a datetime")

        if self.status == ComponentStatus.INITIALIZED:
            self.logger.debug("connect: ping phase")
            for _, inp in self.inputs.items():
                inp.ping()
            self.status = ComponentStatus.CONNECTING
        else:
            self.logger.debug("connect")
            self._connect(start_time)

    def _connect(self, start_time):
        """Connect exchange data and metadata with linked components.

        Components must overwrite this method.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`
            The composition's starting time.
            Can be before the component's actual time.

            Should be passed to :meth:`.try_connect` calls.
        """
        raise NotImplementedError(
            f"Method `_connect` must be implemented by all components, but implementation is missing in {self.name}."
        )

    @final
    def validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have :attr:`.status` :attr:`.ComponentStatus.VALIDATED`.
        """
        self.logger.debug("validate")
        self._validate()
        if self.status != ComponentStatus.FAILED:
            self.status = ComponentStatus.VALIDATED

    def _validate(self):
        """Validate the correctness of the component's settings and coupling.

        Components should overwrite this method.
        """
        self.logger.debug("Method `_validate` not implemented by user.")

    @final
    def update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have :attr:`.status`
        :attr:`.ComponentStatus.UPDATED` or :attr:`.ComponentStatus.FINISHED`.
        """
        if isinstance(self, ITimeComponent):
            self.logger.debug("update - current time: %s", self.time)
        else:
            self.logger.debug("update")

        self._update()

        if self.status not in (ComponentStatus.FAILED, ComponentStatus.FINALIZED):
            self.status = ComponentStatus.UPDATED

    def _update(self):
        """Update the component by one time step.
        Push new values to outputs.

        Components must overwrite this method.
        """
        raise NotImplementedError(
            f"Method `_update` must be implemented by all components, but implementation is missing in {self.name}."
        )

    @final
    def finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have :attr:`.status` :attr:`.ComponentStatus.FINALIZED`.
        """
        self.logger.debug("finalize")
        self._finalize()

        for _n, out in self.outputs.items():
            out.finalize()

        if self.status != ComponentStatus.FAILED:
            self.status = ComponentStatus.FINALIZED

    def _finalize(self):
        """Finalize and clean up the component.

        Components should overwrite this method.
        """
        self.logger.debug("Method `_finalize` not implemented by user.")

    @property
    def inputs(self):
        """dict: The component's inputs."""
        return self._inputs

    @property
    def outputs(self):
        """dict: The component's outputs."""
        return self._outputs

    @property
    def status(self):
        """The component's current status."""
        return self._status

    @status.setter
    def status(self, status):
        """The component's current status."""
        self._status = status

    @property
    def name(self):
        """Component name."""
        return self._name

    @property
    def metadata(self):
        """
        The component's metadata.
        Will only be called after the connect phase from :attr:`Composition.metadata`.

        Components can overwrite this property to add their own specific metadata:

        .. testcode:: metadata

            import finam as fm

            class MyComponent(fm.Component):

                @property
                def metadata(self):
                    # Get the default metadata
                    md = super().metadata

                    # Add your own metadata
                    md["my_field"] = "some value"

                    # Return the dictionary
                    return md

        .. testcode:: metadata
            :hide:

            comp = MyComponent()
            md = comp.metadata

        Returns
        -------
        dict
            A ``dict`` with the following default metadata:
              - ``name`` - the component's name
              - ``class`` - the component's class
              - ``inputs`` - ``dict`` of metadata for all inputs
              - ``outputs`` - ``dict`` of metadata for all outputs
        """
        inputs = {}
        outputs = {}

        for name, inp in self.inputs.items():
            inputs[name] = {
                "name": name,
                "class": inp.__class__.__module__ + "." + inp.__class__.__qualname__,
                "is_static": inp.is_static,
                "info": inp.info.as_dict(),
            }

        for name, out in self.outputs.items():
            outputs[name] = {
                "name": name,
                "class": out.__class__.__module__ + "." + out.__class__.__qualname__,
                "is_static": out.is_static,
                "has_targets": out.has_targets,
                "info": out.info.as_dict(),
            }

        return {
            "name": self.name,
            "class": self.__class__.__module__ + "." + self.__class__.__qualname__,
            "inputs": inputs,
            "outputs": outputs,
        }

    @property
    def logger_name(self):
        """Logger name derived from base logger name and class name."""
        base_logger = logging.getLogger(self.base_logger_name)
        # logger hierarchy indicated by "." in name
        return ".".join(([base_logger.name, self.name]))

    @property
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute. True."""
        return True

    @property
    def connector(self):
        """The component's :class:`.tools.ConnectHelper`.

        See also :meth:`.create_connector` and :meth:`.try_connect`.
        """
        return self._connector

    def create_connector(
        self, pull_data=None, in_info_rules=None, out_info_rules=None, cache=True
    ):
        """Initialize the component's :class:`.tools.ConnectHelper`.

        See also :meth:`.try_connect`, :attr:`.connector` and :class:`.ConnectHelper` for details.

        Parameters
        ----------
        pull_data : arraylike
            Names of the inputs that are to be pulled.
        in_info_rules : dict
            Info transfer rules for inputs. See the examples for details.

            See also :class:`.tools.FromInput`, :class:`.tools.FromOutput` and :class:`.tools.FromValue`.
        out_info_rules : dict
            Info transfer rules for outputs. See the examples for details.

            See also :class:`.tools.FromInput`, :class:`.tools.FromOutput` and :class:`.tools.FromValue`.
        cache : bool
            Whether data and :class:`.Info` objects passed via :meth:`try_connect() <.Component.try_connect>`
            are cached for later calls. Default ``True``.

        Examples
        --------

        The following examples show the usage of this method in :meth:`._initialize`.

        .. testsetup:: *

            import finam as fm
            import datetime as dt

            self = fm.components.CallbackComponent(
                inputs={},
                outputs={},
                callback=lambda inp, _t: {},
                start=dt.datetime(2000, 1, 1),
                step=dt.timedelta(days=1),
            )

        Simple usage if no input data or any metadata from connected components is required:

        .. testcode:: create-connector-simple

            self.inputs.add(name="In", time=self.time, grid=fm.NoGrid())
            self.outputs.add(name="Out", time=self.time, grid=fm.NoGrid())
            self.create_connector()

        To pull specific inputs, use ``pull_data`` like this:

        .. testcode:: create-connector-pull

            self.inputs.add(name="In1", time=self.time, grid=fm.NoGrid())
            self.inputs.add(name="In2", time=self.time, grid=fm.NoGrid())

            self.create_connector(pull_data=["In1", "In2"])

        With the ``in_info_rules`` and ``out_info_rules``, metadata can be transferred between coupling slots.

        Here, the metadata for an output is taken from an input:

        .. testcode:: create-connector-in-to-out

            self.inputs.add(name="In", time=self.time, grid=None, units=None)
            self.outputs.add(name="Out")

            self.create_connector(
                out_info_rules={
                    "Out": [
                        fm.tools.FromInput("In")
                    ]
                }
            )

        The :class:`.Info` object for output ``Out`` will be created and pushed automatically in :meth:`.try_connect`
        as soon as the metadata for ``In`` becomes available.

        Here, the metadata of an output is composed from the metadata of two inputs and a user-defined value:

        .. testcode:: create-connector-in-to-out-multi

            self.inputs.add(name="In1", time=self.time, grid=None, units=None)
            self.inputs.add(name="In2", time=self.time, grid=None, units=None)
            self.outputs.add(name="Out")

            self.create_connector(
                out_info_rules={
                    "Out": [
                        fm.tools.FromInput("In1", ["time", "grid"]),
                        fm.tools.FromInput("In2", ["units"]),
                        fm.tools.FromValue("source", "FINAM"),
                    ]
                }
            )

        The :class:`.Info` object for output ``Out`` would be automatically composed in :meth:`.try_connect`
        as soon as the infos of both inputs become available.
        ``time`` and ``grid`` would be taken from ``In1``, ``units`` from ``In2``,
        and ``source`` would be set to ``"finam"``.

        Rules are evaluated in the given order. Later rules can overwrite attributes set by earlier rules.
        """
        self.logger.trace("create connector")
        self._connector = ConnectHelper(
            self.logger_name,
            self.inputs,
            self.outputs,
            pull_data=pull_data,
            in_info_rules=in_info_rules,
            out_info_rules=out_info_rules,
            cache=cache,
        )
        self.inputs.frozen = True
        self.outputs.frozen = True

    def try_connect(
        self, start_time, exchange_infos=None, push_infos=None, push_data=None
    ):
        """Exchange the info and data with linked components.

        Values passed by the arguments are cached internally for later calls to the method
        if the connector was created with ``cache=True`` (the default).
        Thus, it is sufficient to provide only data and infos that became newly available.
        Giving the same data or infos repeatedly overwrites the cache.

        Sets the component's :attr:`.status` according to success of exchange.

        See also :meth:`.create_connector`, :attr:`.connector` and :class:`.ConnectHelper` for details.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`
            the composition's starting time as passed to :meth:`.Component.try_connect`
        exchange_infos : dict of [str, Info]
            currently or newly available input data infos by input name
        push_infos : dict of [str, Info]
            currently or newly available output data infos by output name
        push_data : dict of [str, array-like]
            currently or newly available output data by output name
        """
        self.logger.trace("try connect")

        if self._connector is None:
            raise FinamStatusError(
                f"No connector in component {self.name}. Call `create_connector()` in `_initialize()`."
            )

        self.status = self._connector.connect(
            start_time,
            exchange_infos=exchange_infos,
            push_infos=push_infos,
            push_data=push_data,
        )
        self.logger.trace("try_connect status is %s", self.status)

    def __getitem__(self, name):
        """Get an input or output by name. Implements access through square brackets.

        Allows for the use of ``comp["Name"]`` as shortcut for ``comp.inputs["Name"]`` and ``comp.outputs["Name"]``.

        Requires that the name does not appear in inputs as well as outputs.

        Returns
        -------
        :class:`.IInput` or :class:`.IOutput`
            The slot with the given name

        Raises
        ------
        KeyError
            If the name occurs in the inputs as well as the outputs,
            or neither in the inputs nor the outputs.
        """
        if name in self.inputs:
            if name in self.outputs:
                msg = f"Name `{name}` exists in inputs as well as outputs of component {self.name}"
                if self.status == ComponentStatus.CREATED:
                    raise KeyError(msg)

                with ErrorLogger(self.logger):
                    raise KeyError(msg)

            return self.inputs[name]

        if name in self.outputs:
            return self.outputs[name]

        msg = f"Name `{name}` does not exist in inputs or outputs of component `{self.name}`"
        if self.status == ComponentStatus.CREATED:
            msg += " The component is not initialized. Did you miss to add it to the composition?"
            raise KeyError(msg)

        with ErrorLogger(self.logger):
            raise KeyError(msg)

    def __repr__(self):
        return self.name


class TimeComponent(ITimeComponent, Component, ABC):
    """Abstract component with time step implementation.

    Extend this class for components with time step.
    See :doc:`/finam-book/development/components`.
    For components without a time step, use :class:`.Component`.

    Derived classes overwrite these methods

    * :meth:`._initialize`
    * :meth:`._connect`
    * :meth:`._validate`
    * :meth:`._update`
    * :meth:`._finalize`
    * :meth:`._next_time`
    """

    def __init__(self):
        Component.__init__(self)
        self._time = None

    @property
    def time(self):
        """The component's current simulation time."""
        if self._time is None and self.status in (
            ComponentStatus.CREATED,
            ComponentStatus.INITIALIZED,
        ):
            return None

        if not isinstance(self._time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")
        return self._time

    @time.setter
    def time(self, time):
        if not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")
        self._time = time

    @property
    def next_time(self):
        """The component's predicted simulation time of its next pulls.

        Can be ``None`` if the component has no inputs.
        """
        return self._next_time()

    def _next_time(self):
        raise NotImplementedError(
            "Method `_next_time` must be implemented by all time components, "
            f"but implementation is missing in {self.name}."
        )


class IOType(IntEnum):
    """IOType of the IOList."""

    INPUT = 0
    OUTPUT = 1


class IOList(collections.abc.Mapping):
    """
    Map for IO.

    Parameters
    ----------
    owner : :class:`.IComponent`
        The owning component of this IOList
    io_type : int, str, IOType
        IO type. Either "INPUT" or "OUTPUT".
    """

    def __init__(self, owner, io_type):
        """
        _summary_

        Parameters
        ----------
        io_type : _type_
            _description_
        """
        self.owner = owner
        self.type = get_enum_value(io_type, IOType)
        self.cls = [Input, Output][self.type]
        self.name = self.cls.__name__
        self.icls = [IInput, IOutput][self.type]
        self.iname = self.icls.__name__
        self._dict = {}
        self.frozen = False

    def add(self, io=None, *, name=None, info=None, static=False, **info_kwargs):
        """
        Add a new IO object either directly ob by attributes.

        Parameters
        ----------
        io : :class:`.IInput` or :class:`.IOutput`, optional
            IO object to add, by default None
        name : str, optional
            Name of the new IO object to add, by default None
        info : :class:`.Info`, optional
            Info of the new IO object to add, by default None
        static : bool, optional
            Whether the new IO object in static, by default False
        **info_kwargs
            Optional keyword arguments to instantiate an Info object

        Raises
        ------
        ValueError
            If io is not of the correct type.
        """
        if self.frozen:
            raise ValueError("IO.add: list is frozen.")
        io = (
            self.cls(name=name, info=info, static=static, **info_kwargs)
            if io is None
            else io
        )
        if not isinstance(io, self.icls):
            raise ValueError(f"IO.add: {self.name} is not of type {self.iname}")
        if io.name in self._dict:
            raise ValueError(f"IO.add: {self.name} '{io.name}' already exists.")
        self._dict[io.name] = io

    @property
    def names(self):
        """list: all IO names in this list."""
        return list(self)

    def set_logger(self, module):
        """
        Set the logger in the items of the IOList.

        Parameters
        ----------
        module : :class:`.IComponent`
            Module holding the IOList.

        Raises
        ------
        FinamLogError
            When item is loggable but not the base module.
        """
        for name, item in self.items():
            if (
                is_loggable(item)
                and item.uses_base_logger_name
                and not is_loggable(module)
            ):
                mname = getattr(module, "name", None)
                raise FinamLogError(
                    f"IO: {self.name} '{name}' can't get logger from '{mname}'."
                )
            if is_loggable(item) and item.uses_base_logger_name:
                item.base_logger_name = module.logger_name

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __contains__(self, item):
        return item in self._dict

    def __getitem__(self, key):
        """Access an item by name."""
        if key in self._dict:
            return self._dict[key]

        if self.owner is None:
            raise KeyError(f"No {self.cls.__name__} `{key}` in unknown component.")

        msg = f"No {self.cls.__name__} `{key}` in component `{self.owner.name}`."
        if self.owner.status == ComponentStatus.CREATED:
            msg += " The component is not initialized. Did you miss to add it to the composition?"
            raise KeyError(msg)

        with ErrorLogger(self.owner.logger):
            raise KeyError(msg)

    def __setitem__(self, key, value):
        if self.frozen:
            raise ValueError("IO: list is frozen.")
        if key in self._dict:
            raise ValueError(f"IO: {self.name} '{key}' already exists.")
        if not isinstance(value, self.icls):
            raise ValueError(f"IO: {self.name} is not of type {self.iname}")
        if key != value.name:
            raise ValueError(
                f"IO: {self.name} name '{value.name}' differs from key '{key}'"
            )
        self._dict[key] = value

    def __str__(self):
        return str(self._dict)

    def __repr__(self):
        return repr(self._dict)



import unittest
from datetime import datetime, timedelta
from functools import partial

import finam as fm
from finam.components.debug import DebugConsumer


class PullComponent(fm.Component):
    def __init__(self):
        super().__init__()

    def _initialize(self):
        self.outputs.add(
            fm.CallbackOutput(
                callback=partial(self._get_data, "Out"),
                name="Out",
                time=None,
                grid=fm.NoGrid(),
            )
        )
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _get_data(self, _name, _caller, time):
        return time.day


class TestPullBasedComponent(unittest.TestCase):
    def test_component(self):
        time = datetime(2000, 1, 1)

        pull_comp = PullComponent()
        consumer = DebugConsumer(
            inputs={"In": fm.Info(time=time, grid=fm.NoGrid())},
            start=time,
            step=timedelta(days=1),
        )

        composition = fm.Composition([pull_comp, consumer], log_level="DEBUG")

        pull_comp.outputs["Out"] >> consumer.inputs["In"]

        composition.run(start_time=time, end_time=datetime(2000, 1, 12))

        self.assertEqual(consumer.data, {"In": 12})


if __name__ == "__main__":
    unittest.main()



"""Mask tools for FINAM."""

from enum import Enum

import numpy as np

from ...errors import FinamDataError
from .units import UNITS, is_quantified, quantify

MASK_INDICATORS = ["_FillValue", "missing_value"]


class Mask(Enum):
    """Mask settings for Info."""

    FLEX = 0
    """Data can be masked or unmasked."""
    NONE = 1
    """Data is expected to be unmasked and given as plain numpy arrays."""


def is_masked_array(data):
    """
    Check if data is a masked array.

    Parameters
    ----------
    data : Any
        The given data array.

    Returns
    -------
    bool
        Whether the data is a MaskedArray.
    """
    if is_quantified(data):
        return np.ma.isMaskedArray(data.magnitude)
    return np.ma.isMaskedArray(data)


def has_masked_values(data):
    """
    Determine whether the data has masked values.

    Parameters
    ----------
    data : Any
        The given data array.

    Returns
    -------
    bool
        Whether the data is a MaskedArray and has any masked values.
    """
    return np.ma.is_masked(data)


def filled(data, fill_value=None):
    """
    Return input as an array with masked data replaced by a fill value.

    This routine respects quantified and un-quantified data.

    Parameters
    ----------
    data : :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        The reference object input.
    fill_value : array_like, optional
        The value to use for invalid entries. Can be scalar or non-scalar.
        If non-scalar, the resulting ndarray must be broadcastable over
        input array. Default is None, in which case, the `fill_value`
        attribute of the array is used instead.

    Returns
    -------
    pint.Quantity or numpy.ndarray
        New object with the same shape and type as other,
        with the data filled with fill_value.
        Units will be taken from the input if present.

    See also
    --------
    :func:`numpy.ma.filled`:
        Numpy routine doing the same.
    """
    if not is_masked_array(data):
        return data
    if is_quantified(data):
        return UNITS.Quantity(data.magnitude.filled(fill_value), data.units)
    return data.filled(fill_value)


def to_masked(data, **kwargs):
    """
    Return a masked version of the data.

    Parameters
    ----------
    data : :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        The reference object input.
    **kwargs
        keyword arguments forwarded to :any:`numpy.ma.array`

    Returns
    -------
    pint.Quantity or numpy.ma.MaskedArray
        New object with the same shape and type but as a masked array.
        Units will be taken from the input if present.
    """
    if is_masked_array(data) and not kwargs:
        return data
    if is_quantified(data):
        return UNITS.Quantity(np.ma.array(data.magnitude, **kwargs), data.units)
    return np.ma.array(data, **kwargs)


def to_compressed(xdata, order="C", mask=None):
    """
    Return all the non-masked data as a 1-D array respecting the given array order.

    Parameters
    ----------
    data : :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        The reference object input.
    order : str
        order argument for :any:`numpy.ravel`
    mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray`, optional
        mask to use when data is not masked already

    Returns
    -------
    :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        New object with the flat shape and only unmasked data but and same type as input.
        Units will be taken from the input if present.

    See also
    --------
    :func:`numpy.ma.compressed`:
        Numpy routine doing the same but only for C-order.
    """
    is_masked = is_masked_array(xdata)
    if is_masked or (mask is not None and mask_specified(mask)):
        data = np.ravel(xdata.data if is_masked else xdata, order)
        mask = xdata.mask if is_masked else mask
        if mask is not np.ma.nomask:
            data = data.compress(np.logical_not(np.ravel(mask, order)))
        return quantify(data, xdata.units) if is_quantified(xdata) else data
    return np.reshape(xdata, -1, order=order)


def from_compressed(xdata, shape, order="C", mask=None, **kwargs):
    """
    Fill a (masked) array following a given mask or shape with the provided data.

    This will only create a masked array if kwargs are given (especially a mask).
    Otherwise this is simply reshaping the given data.
    Filling is performed in the given array order.

    Parameters
    ----------
    data : :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        The reference object input.
    shape : str
        shape argument for :any:`numpy.reshape`
    order : str
        order argument for :any:`numpy.reshape`
    mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray`
        mask to use
    **kwargs
        keyword arguments forwarded to :any:`numpy.ma.array`

    Returns
    -------
    :class:`pint.Quantity` or :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
        New object with the desired shape and same type as input.
        Units will be taken from the input if present.
        Will only be a masked array if kwargs are given.

    See also
    --------
    to_compressed:
        Inverse operation.
    :any:`numpy.ma.array`:
        Routine consuming kwargs to create a masked array.
    :any:`numpy.reshape`:
        Equivalent routine if no mask is provided.

    Notes
    -----
    If both `mask` and `shape` are given, they need to match in size.
    """
    if mask is None or mask is np.ma.nomask or not mask_specified(mask):
        if kwargs and mask is Mask.NONE:
            msg = "from_compressed: Can't create masked array with mask=Mask.NONE"
            raise FinamDataError(msg)
        data = np.reshape(xdata, shape, order=order)
        return to_masked(data, **kwargs) if kwargs or mask is np.ma.nomask else data
    if is_quantified(xdata):
        # pylint: disable-next=unexpected-keyword-arg
        data = quantify(np.empty_like(xdata, shape=np.prod(shape)), xdata.units)
    else:
        # pylint: disable-next=unexpected-keyword-arg
        data = np.empty_like(xdata, shape=np.prod(shape))
    data[np.logical_not(np.ravel(mask, order=order))] = xdata
    return to_masked(np.reshape(data, shape, order=order), mask=mask, **kwargs)


def check_data_covers_domain(data, mask=None):
    """
    Check if the given data covers a domain defined by a mask on the same grid.

    Parameters
    ----------
    data : Any
        The given data array for a single time-step.
    mask : None or bool or array of bool, optional
        Mask defining the target domain on the same grid as the data,
        by default None

    Returns
    -------
    bool
        Whether the data covers the desired domain.

    Raises
    ------
    ValueError
        When mask is given and mask and data don't share the same shape.
    """
    if not _is_single_mask_value(mask) and np.shape(mask) != np.shape(data):
        raise ValueError("check_data_covers_domain: mask and data shape differ.")
    if not has_masked_values(data):
        return True
    if _is_single_mask_value(mask):
        return bool(mask)
    return np.all(mask[data.mask])


def _is_single_mask_value(mask):
    return mask is None or mask is np.ma.nomask or mask is False or mask is True


def masks_compatible(
    this, incoming, incoming_donwstream, this_grid=None, incoming_grid=None
):
    """
    Check if an incoming mask is compatible with a given mask.

    Parameters
    ----------
    this : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None
        mask specification to check against
    incoming : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None
        incoming mask to check for compatibility
    incoming_donwstream : bool
        Whether the incoming mask is from downstream data
    this_grid : Grid or NoGrid or None, optional
        grid for first mask (to check shape and value equality)
    incoming_grid : Grid or NoGrid or None, optional
        grid for second mask (to check shape and value equality)

    Returns
    -------
    bool
        mask compatibility
    """
    if incoming_donwstream:
        upstream, downstream = this, incoming
        up_grid, down_grid = this_grid, incoming_grid
    else:
        upstream, downstream = incoming, this
        up_grid, down_grid = incoming_grid, this_grid
    # None is incompatible
    if upstream is None:
        return False
    # Mask.FLEX accepts anything, Mask.NONE only Mask.NONE
    if not mask_specified(downstream):
        if not mask_specified(upstream):
            return downstream == Mask.FLEX or upstream == Mask.NONE
        return downstream == Mask.FLEX
    # if mask is specified, upstream mask must also be specified
    if not mask_specified(upstream):
        return False
    # if both mask given, compare them
    return masks_equal(downstream, upstream, down_grid, up_grid)


def masks_equal(this, other, this_grid=None, other_grid=None):
    """
    Check two masks for equality.

    Parameters
    ----------
    this : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None
        first mask
    other : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None
        second mask
    this_grid : Grid or NoGrid or None, optional
        grid for first mask (to check shape and value equality)
    other_grid : Grid or NoGrid or None, optional
        grid for second mask (to check shape and value equality)

    Returns
    -------
    bool
        mask equality
    """
    if this is None and other is None:
        return True
    if not mask_specified(this) and not mask_specified(other):
        return this == other
    # need a valid mask at this point
    if not np.ma.is_mask(this) or not np.ma.is_mask(other):
        return False
    # special treatment of "nomask"
    if this is np.ma.nomask:
        if other is np.ma.nomask:
            return True
        return not np.any(other)
    if other is np.ma.nomask:
        return not np.any(this)
    # compare masks
    if not np.ndim(this) == np.ndim(other):
        return False
    # mask shape is grid specific (reversed axes, decreasing axis)
    if this_grid is None or other_grid is None:
        return True
    this = this_grid.to_canonical(this)
    other = other_grid.to_canonical(other)
    if not np.all(np.shape(this) == np.shape(other)):
        return False
    return np.all(this == other)


def is_sub_mask(mask, submask):
    """
    Check for a sub-mask.

    Parameters
    ----------
    mask : arraylike
        The original mask.
    submask : arraylike
        The potential submask.

    Returns
    -------
    bool
        Whether 'submask' is a sub-mask of 'mask'.
    """
    if not np.ma.is_mask(mask) or not np.ma.is_mask(submask):
        return False
    if mask is np.ma.nomask:
        return True
    if submask is np.ma.nomask:
        return not np.any(mask)
    if not np.ndim(mask) == np.ndim(submask):
        return False
    if not np.all(np.shape(mask) == np.shape(submask)):
        return False
    return np.all(submask[mask])


def mask_specified(mask):
    """
    Determine whether given mask selection indicates a masked array.

    Parameters
    ----------
    mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray`
        mask to check

    Returns
    -------
    bool
        False if mask is Mask.FLEX or Mask.NONE, True otherwise
    """
    return not any(mask is val for val in list(Mask))



"""Parametric grid generator components"""
import numpy as np

from finam.data.grid_base import Grid
from finam.data.grid_spec import NoGrid, StructuredGrid
from finam.data.tools import Info
from finam.errors import FinamMetaDataError
from finam.interfaces import ComponentStatus
from finam.sdk import CallbackOutput, Component, Output
from finam.tools import ErrorLogger


class ParametricGrid(Component):
    """Pull-based parametric grid generator.

    Generates grids with values filled from a function of time and cell coordinates.

    .. code-block:: text

        +----------------+
        |                |
        | ParametricGrid | [Grid] -->
        |                |
        +----------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        component = fm.components.ParametricGrid(
            info=fm.Info(time=None, grid=fm.UniformGrid((20, 15))),
            func=lambda t, x, y: x * y,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    info : Info, optional
        Output metadata info. All values are taken from the target if not specified.
    func : callable
        A callable with the signature func(t, x[, y[, z]]) -> Quantity
    """

    def __init__(
        self,
        info=None,
        func=None,
    ):
        super().__init__()

        self._info = info or Info(time=None, grid=None, units=None)
        self._func = func
        self._is_ready = False

    def _initialize(self):
        self.outputs.add(
            io=CallbackOutput(
                callback=self._generate_grid, name="Grid", info=self._info
            )
        )
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)

        info = self.connector.out_infos["Grid"]
        if info is not None:
            failed = isinstance(info.grid, NoGrid)
            failed |= isinstance(info.grid, Grid) and not 1 <= info.grid.dim <= 3
            if failed:
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError(
                        f"Can generate parametric grids only for gridded 1D-3D data. "
                        f"Got {info.grid}"
                    )

            self._info = info
            self._is_ready = True

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _generate_grid(self, _caller, time):
        if not self._is_ready:
            return None

        return _generate_grid(self._info.grid, time, self._func)


class StaticParametricGrid(Component):
    """Static parametric grid generator.

    Generates a grid with values filled from a function of cell coordinates.

    .. code-block:: text

        +----------------------+
        |                      |
        | StaticParametricGrid | [Grid] -->
        |                      |
        +----------------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        component = fm.components.StaticParametricGrid(
            info=fm.Info(time=None, grid=fm.UniformGrid((20, 15))),
            func=lambda x, y: x * y,
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    info : Info, optional
        Output metadata info. All values are taken from the target if not specified.
    func : callable
        A callable with the signature func(x[, y[, z]]) -> Quantity
    """

    def __init__(
        self,
        info=None,
        func=None,
    ):
        super().__init__()

        self._info = info or Info(time=None, grid=None, units=None)
        self._func = func

        self._is_ready = False

    def _initialize(self):
        self.outputs.add(io=Output(name="Grid", static=True, info=self._info))
        self.create_connector()

    def _connect(self, start_time):
        push_data = {}
        if self.connector.out_infos["Grid"] is not None:
            push_data["Grid"] = self._generate_grid()

        self.try_connect(start_time, push_data=push_data)

        info = self.connector.out_infos["Grid"]
        if info is not None:
            failed = isinstance(info.grid, NoGrid)
            failed |= isinstance(info.grid, Grid) and not 1 <= info.grid.dim <= 3
            if failed:
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError(
                        f"Can generate parametric grid only for gridded 1D-3D data. "
                        f"Got {info.grid}"
                    )

            self._info = info
            self._is_ready = True

    def _validate(self):
        self.status = ComponentStatus.FINISHED

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _generate_grid(self):
        return _generate_grid(self._info.grid, None, self._func)


def _generate_grid(grid, time, cell_func):
    points = grid.data_points
    data = np.full((points.shape[0],), 0.0, dtype=float)

    if time is None:
        for i, p in enumerate(points):
            data[i] = cell_func(*p)
    else:
        for i, p in enumerate(points):
            data[i] = cell_func(time, *p)

    if isinstance(grid, StructuredGrid):
        data = np.reshape(data, newshape=grid.data_shape, order=grid.order)

    return data



"""
Abstract base implementation for adapters.
"""

import logging
from abc import ABC
from datetime import datetime
from typing import final

from ..data import tools
from ..data.tools import Info
from ..errors import FinamLogError, FinamMetaDataError, FinamTimeError
from ..interfaces import IAdapter, IOutput, ITimeDelayAdapter
from ..tools.log_helper import ErrorLogger, is_loggable
from .input import Input
from .output import Output


class Adapter(IAdapter, Input, Output, ABC):
    """Abstract adapter implementation.

    Extend this class for adapters.
    See :doc:`/finam-book/development/adapters`.

    Simple derived classes overwrite :meth:`._get_data`.

    Adapters that alter the metadata can intercept it in :meth:`._get_info`

    For time-dependent adapters with push-functionality, also overwrite the following:

    * :meth:`._source_updated`
    * :attr:`.needs_push`
    """

    def __init__(self):
        Input.__init__(self, name=self.__class__.__name__)
        Output.__init__(self, name=self.__class__.__name__)
        self._name = self.__class__.__name__
        self._source = None
        self._targets = []

    def with_name(self, name):
        """Renames the adapter and returns self."""
        self._name = name
        return self

    @property
    def time(self):
        """The output's time of the latest available data"""
        return None

    @property
    def is_static(self):
        return False

    @final
    @property
    def info(self):
        return self._output_info

    @final
    @property
    def in_info(self):
        """Info from connected source."""
        return self._input_info

    @property
    def needs_pull(self):
        """bool: if the adapter needs pull."""
        return False

    @property
    def needs_push(self):
        """bool: if the adapter needs push."""
        return False

    @property
    def metadata(self):
        """
        The adapter's metadata.
        Will only be called after the connect phase from :attr:`Composition.metadata`.

        Adapters can overwrite this property to add their own specific metadata:

        .. testcode:: metadata

            import finam as fm

            class MyAdapter(fm.Adapter):

                @property
                def metadata(self):
                    # Get the default metadata
                    md = super().metadata

                    # Add your own metadata
                    md["my_field"] = "some value"

                    # Return the dictionary
                    return md

        .. testcode:: metadata
            :hide:

            ada = MyAdapter()


        Returns
        -------
        dict
            A ``dict`` with the following default metadata:
              - ``name`` - the component's name
              - ``class`` - the component's class
        """
        meta = {
            "name": self.name,
            "class": self.__class__.__module__ + "." + self.__class__.__qualname__,
            "out_info": self._output_info.as_dict(),
        }

        if self._input_info is not None:
            meta["in_info"] = self._input_info.as_dict()

        return meta

    @final
    def push_data(self, data, time):
        """Push data into the output.

        Should notify targets, and can handle the provided date.

        Parameters
        ----------
        data : array_like
            Data set to push.
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the data set.
        """
        self.logger.debug("push data")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")

        self.notify_targets(time)

    @final
    def source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the notification.
        """
        self.logger.debug("source updated")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")

        self._source_updated(time)

        self.notify_targets(time)

    def notify_targets(self, time):
        """Notify all targets by calling their ``source_updated(time)`` method.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the simulation.
        """
        self.logger.trace("notify targets")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise ValueError("Time must be of type datetime")

        for target in self.targets:
            target.source_updated(time)

    def _source_updated(self, time):
        """Informs the input that a new output is available.

        Adapters can overwrite this method to handle incoming data.

        Adapters that make use of this method to accumulate data should also
        overwrite :attr:`.needs_push` to return ``True``.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time of the notification.
        """

    def get_data(self, time, target):
        """Get the transformed data of this adapter.

        Internally calls :meth:`._get_data`.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput`
            Requesting end point of this pull.

        Returns
        -------
        :class:`pint.Quantity`
            Transformed data-set for the requested time.
        """
        self.logger.debug("get data")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise FinamTimeError("Time must be of type datetime")

        data = self._get_data(time, target)

        with ErrorLogger(self.logger):
            xdata, conv = tools.prepare(data, self._output_info, report_conversion=True)
            if conv is not None:
                self.logger.profile(
                    "converted units from %s to %s (%d entries)", *conv, xdata.size
                )
            return xdata

    def _get_data(self, time, target):
        """Get the transformed data of this adapter.

        Adapters must overwrite this method.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput`
            Requesting end point of this pull.

        Returns
        -------
        :class:`pint.Quantity`
            Transformed data-set for the requested time.
        """
        raise NotImplementedError(
            f"Method `_get_data` must be implemented by all adapters, but implementation is missing in {self.name}."
        )

    def get_info(self, info):
        """Exchange and get the output's data info.

        Parameters
        ----------
        info : :class:`.Info`
            Requested data info

        Returns
        -------
        Info
            Delivered data info

        Raises
        ------
        FinamNoDataError
            Raises the error if no info is available
        """
        self.logger.trace("get info")
        self._output_info = self._get_info(info)
        return self._output_info

    def _get_info(self, info):
        """Exchange and get the output's data info.

        Adapters can overwrite this method to manipulate the metadata for the output.

        Parameters
        ----------
        info : :class:`.Info`
            Requested data info

        Returns
        -------
        Info
            Delivered data info
        """
        return self.exchange_info(info)

    def pinged(self, source):
        """Called when receiving a ping from a downstream input."""
        self._source.pinged(self if self.needs_push else source)

    @final
    def exchange_info(self, info=None):
        """Exchange the data info with the input's source.

        Parameters
        ----------
        info : :class:`.Info`
            request parameters

        Returns
        -------
        Info
            delivered parameters
        """
        self.logger.trace("exchanging info")
        with ErrorLogger(self.logger):
            if info is None:
                raise FinamMetaDataError("No metadata provided")
            if not isinstance(info, Info):
                raise FinamMetaDataError("Metadata must be of type Info")

        in_info = self._source.get_info(info)

        self._input_info = in_info
        self._output_info = in_info
        return in_info

    @property
    def source(self):
        """Get the input's source output or adapter

        Returns
        -------
        :class:`.IOutput`
            The input's source.
        """
        return super().source

    @source.setter
    def source(self, source):
        """Set the adapter input's source output or adapter

        Parameters
        ----------
        source :
            source output or adapter
        """
        self.logger.trace("set source")
        # fix to set base-logger for adapters derived from Input source logger
        if self.uses_base_logger_name and not is_loggable(source):
            with ErrorLogger(self.logger):
                raise FinamLogError(
                    f"Adapter '{self.name}' can't get base logger from its source."
                )
        else:
            self.base_logger_name = source.logger_name

        with ErrorLogger(self.logger):
            if self._source is not None:
                raise ValueError(
                    "Source of input is already set! "
                    "(You probably tried to connect multiple outputs to a single input)"
                )
            if not isinstance(source, IOutput):
                raise ValueError("Only IOutput can be set as source for Input")

        self._source = source

    @property
    def logger_name(self):
        """Logger name derived from source logger name and class name."""
        base_logger = logging.getLogger(self.base_logger_name)
        return ".".join(([base_logger.name, " >> ", self.name]))

    def finalize(self):
        """Called at the end of each run. Calls :meth:`._finalize`."""
        self.logger.debug("finalize")
        self._finalize()

    def _finalize(self):
        """Called at the end of each run. Overwrite this for cleanup."""


class TimeDelayAdapter(Adapter, ITimeDelayAdapter, ABC):
    """Base class for adapters that delay/offset time to resolve dependency cycles."""

    def __init__(self):
        super().__init__()
        self.initial_time = None

    def get_data(self, time, target):
        """Get the transformed data of this adapter.

        Internally calls :meth:`._get_data`.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            Simulation time to get the data for.
        target : :class:`.IInput`
            Requesting end point of this pull.

        Returns
        -------
        :class:`pint.Quantity`
            Transformed data-set for the requested time.
        """
        self.logger.debug("get data")
        if time is not None and not isinstance(time, datetime):
            with ErrorLogger(self.logger):
                raise FinamTimeError("Time must be of type datetime")

        new_time = self.with_delay(time)
        data = self._get_data(new_time, target)

        self._pulled(time)

        with ErrorLogger(self.logger):
            xdata, conv = tools.prepare(data, self._output_info, report_conversion=True)
            if conv is not None:
                self.logger.profile(
                    "converted units from %s to %s (%d entries)", *conv, xdata.size
                )
            return xdata

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : :class:`datetime <datetime.datetime>`
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        d = self.pull_data(time, target)
        return d

    def _pulled(self, time):
        """This method is called during pulls, with the original pull time.

        Can be overwritten to store the original pull time,
        as in :meth:`._get_data()` only the manipulated time is available.

        Called after :meth:`._get_data()` (i.e. after the actual pull).

        Parameters
        ----------

        time : :class:`datetime <datetime.datetime>`
            The original (requested) time of the current pull.
        """

    def get_info(self, info):
        """Exchange and get the output's data info.

        Parameters
        ----------
        info : :class:`.Info`
            Requested data info

        Returns
        -------
        dict
            Delivered data info

        Raises
        ------
        FinamNoDataError
            Raises the error if no info is available
        """
        self.logger.trace("get info")
        self._output_info = self._get_info(info)
        self.initial_time = self._output_info.time
        return self._output_info



"""Logging helpers."""
# pylint: disable=E1101
import logging
import sys
from contextlib import AbstractContextManager

from finam.interfaces import Loggable

from . import wurlitzer


def is_loggable(obj):
    """
    Check if given object is loggable.

    Parameters
    ----------
    obj : object
        Object to check for loggability

    Returns
    -------
    bool
        Loggability of the object.
    """
    return isinstance(obj, Loggable)


class LogWriter:
    """
    Log writer.

    Parameters
    ----------
    logger : string, None or logging.Logger instance, optional
        Logger name for this writer. Will be the root logger by default.
    level : integer, optional
        Logging level, by default logging.INFO
    """

    def __init__(self, logger=None, level=logging.INFO):
        self.logger = logger.name if isinstance(logger, logging.Logger) else logger
        self.level = level

    def write(self, msg):
        """
        Write a given message to the logger.

        Parameters
        ----------
        msg : string
            Message to log.
        """
        logger = logging.getLogger(self.logger)
        if msg != "\n":
            logger.log(self.level, msg)


class LogStdOutStdErr(AbstractContextManager):
    """
    Context manager to redirect stdout and stderr to a logger.

    Parameters
    ----------
    logger : string, None or logging.Logger instance, optional
        Logger name for this writer. Will be the root logger by default.

    level_stdout : integer, optional
        Logging level for stdout, by default logging.INFO

    level_stderr : integer, optional
        Logging level for stderr, by default logging.WARN
    """

    def __init__(
        self, logger=None, level_stdout=logging.INFO, level_stderr=logging.WARN
    ):
        self._stdout_target = LogWriter(logger=logger, level=level_stdout)
        self._stderr_target = LogWriter(logger=logger, level=level_stderr)
        self._old_stdout = getattr(sys, "stdout")
        self._old_stderr = getattr(sys, "stderr")

    def __enter__(self):
        setattr(sys, "stdout", self._stdout_target)
        setattr(sys, "stderr", self._stderr_target)

    def __exit__(self, *args, **kwargs):
        setattr(sys, "stdout", self._old_stdout)
        setattr(sys, "stderr", self._old_stderr)


class LogCStdOutStdErr(AbstractContextManager):
    """
    Context manager to redirect low-level C stdout and stderr to a logger.

    Parameters
    ----------
    logger : string, None or logging.Logger instance, optional
        Logger name for this writer. Will be the root logger by default.

    level_stdout : integer, optional
        Logging level for stdout, by default logging.INFO

    level_stderr : integer, optional
        Logging level for stderr, by default logging.WARN
    """

    def __init__(
        self, logger=None, level_stdout=logging.INFO, level_stderr=logging.WARN
    ):
        self.logger = logger.name if isinstance(logger, logging.Logger) else logger
        self.level_stdout = level_stdout
        self.level_stderr = level_stderr
        self.stdout = None
        self.stderr = None
        self.pipes = wurlitzer.pipes()

    def __enter__(self):
        self.stdout, self.stderr = self.pipes.__enter__()

    def __exit__(self, *args, **kwargs):
        self.pipes.__exit__(*args, **kwargs)
        logger = logging.getLogger(self.logger)
        for line in self.stdout.read().splitlines():
            logger.log(self.level_stdout, line)
        for line in self.stderr.read().splitlines():
            logger.log(self.level_stderr, line)


class ErrorLogger(AbstractContextManager):
    """
    Context manager to log Exceptions.

    Parameters
    ----------
    logger : string, None or logging.Logger instance, optional
        Logger name to use. Will be the root logger by default.
    do_log : Bool, optional
        Whether to really log errors. Will be true by default.
    """

    def __init__(self, logger=None, do_log=True):
        self.logger = logger.name if isinstance(logger, logging.Logger) else logger
        self.do_log = do_log

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value is not None and self.do_log:
            logging.getLogger(self.logger).exception(exc_value)


def add_logging_level(name, num, method=None):
    """
    Adds a logging level to the :mod:`logging` module.

    Examples
    --------

    .. code-block:: Python

        add_logging_level("TRACE", logging.DEBUG - 5)

    Parameters
    ----------
    name : str
        The name of the new logging level.
    num : int
        The numeric severity of the new logging level.
    method : str, optional
        The method name for the new logging level.
        Defaults to lowercase of ``name``.
    """
    if not method:
        method = name.lower()

    if hasattr(logging, name):
        raise AttributeError(f"{name} already defined in logging module")
    if hasattr(logging.getLoggerClass(), name):
        raise AttributeError(f"{name} already defined in logger class")

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(num):
            # pylint: disable=protected-access
            self._log(num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(num, message, *args, **kwargs)

    logging.addLevelName(num, name)
    setattr(logging, name, num)
    setattr(logging.getLoggerClass(), method, log_for_level)
    setattr(logging, method, log_to_root)



"""Enum conversion helper."""

from enum import Enum


def get_enum_value(value, enum_cls, error=ValueError):
    """
    Convert value to corresponding enum value.

    Parameters
    ----------
    value : any
        Value to convert to enum value.
    enum_cls : Enum
        Enumeration class to get value from.
    error : Exception, optional
        Error to raise if value is not present in enum class,
        by default ValueError

    Returns
    -------
    enum
        Corresponding Enum value.

    Raises
    ------
    ValueError
        If enum_cls is not a subclass of Enum
    error
        If value couldn't be found in given enum class.
    """
    if not isinstance(enum_cls, type) or not issubclass(enum_cls, Enum):
        raise ValueError(f"Class is not of type Enum: {enum_cls}")
    if value in list(enum_cls) and not issubclass(enum_cls, int):
        return value
    if value in [e.value for e in enum_cls]:
        return enum_cls(value)
    if value in [e.name for e in enum_cls]:
        return enum_cls[value]
    raise error(f"Unknown {enum_cls.__name__} value '{value}'")



"""Pull-based components for merging multiple inputs into a single output"""
from finam.interfaces import ComponentStatus

from ..data.tools import compatible_units, strip_time
from ..errors import FinamMetaDataError
from ..sdk import CallbackOutput, Component
from ..tools.log_helper import ErrorLogger


class WeightedSum(Component):
    """Pull-based component to merge inputs by weighted sum.

    This component does not have an own time step.
    Execution is triggered by downstream pulls.

    .. code-block:: text

                             +-------------------+
      --> [<custom1>       ] |                   |
      --> [<custom1>_weight] |                   |
      --> [<custom2>       ] | CallbackComponent | [WeightedSum] -->
      --> [<custom2>_weight] |                   |
      --> [....            ] |                   |
                             +-------------------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        component = fm.components.WeightedSum(inputs=["A", "B", "C"])

        # ... create and initialize composition

        # comp_1.outputs["Value"] >> component.inputs["A"]
        # comp_1.outputs["Weight"] >> component.inputs["A_weight"]
        # ...

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    inputs : list(str)
        Base input names; will create two inputs for each entry: "<name>" and "<name>_weight"
    start : :class:`datetime <datetime.datetime>`
        Starting time, for initial data exchange
    grid : GridBase or None
        Expected input grid specification; tries to obtain grid specs from inputs if set to None
    """

    def __init__(self, inputs, grid=None):
        super().__init__()
        self._input_names = inputs
        self._grid = grid
        self._units = None
        self._in_data = None
        self._out_data = None
        self._last_update = None

    def _initialize(self):
        for name in self._input_names:
            self.inputs.add(name=name, time=None, grid=self._grid, units=None)
            self.inputs.add(name=name + "_weight", time=None, grid=self._grid, units="")

        self._grid = None

        self.outputs.add(CallbackOutput(callback=self._get_data, name="WeightedSum"))
        self.create_connector(pull_data=list(self.inputs))

    def _connect(self, start_time):
        push_infos = self._check_infos()
        self.try_connect(start_time, push_infos=push_infos)

        if self.status == ComponentStatus.CONNECTED:
            # just to check for all inputs equal
            _push_infos = self._check_infos()

        if self.connector.all_data_pulled:
            self._in_data = self.connector.in_data

    def _check_infos(self):
        push_infos = {}
        for name in self._input_names:
            info = self.connector.in_infos[name]
            if info is not None:
                if not self.connector.infos_pushed["WeightedSum"]:
                    push_infos["WeightedSum"] = info.copy_with()

                self._check_grid(info)
                self._compatible_units(info)

            weight_info = self.connector.in_infos[name + "_weight"]
            if weight_info is not None:
                self._check_grid(weight_info)

        return push_infos

    def _check_grid(self, info):
        if self._grid is None:
            self._grid = info.grid
        else:
            if self._grid != info.grid:
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError("All inputs must have the same grid.")

    def _compatible_units(self, info):
        if self._units is None:
            self._units = info.units
        else:
            if not compatible_units(self._units, info.units):
                with ErrorLogger(self.logger):
                    raise FinamMetaDataError(
                        "All value inputs must have the same dimensions."
                    )

    def _validate(self):
        pass

    def _update(self):
        pass

    def _finalize(self):
        pass

    def _get_data(self, _caller, time):
        if self._in_data is None:
            return None

        if time != self._last_update:
            if self.status == ComponentStatus.VALIDATED:
                self._in_data = {
                    name: inp.pull_data(time) for name, inp in self.inputs.items()
                }

            result = None

            for name in self._input_names:
                value = strip_time(self._in_data[name], self._grid)
                weight = strip_time(self._in_data[name + "_weight"], self._grid)

                if result is None:
                    result = value * weight
                else:
                    result += value * weight

            self._out_data = result
            self._last_update = time

        return self._out_data



"""
Unit tests for data info propagation.
"""

import unittest
from datetime import datetime, timedelta

from finam import Adapter, Composition, FinamMetaDataError, Info, NoGrid, TimeComponent
from finam.components.generators import CallbackGenerator


class MockupConsumer(TimeComponent):
    def __init__(self, time, info):
        super().__init__()
        self.time = time
        self.step = timedelta(days=1)
        self.info = info
        self.data = None

    def _next_time(self):
        return self.time + self.step

    def _initialize(self):
        self.inputs.add(name="Input", info=self.info)
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time)

    def _validate(self):
        pass

    def _update(self):
        self.data = self.inputs["Input"].pull_data(self.time, None)
        self.time += self.step

    def _finalize(self):
        pass


class MockupProducer(TimeComponent):
    def __init__(self, time, info):
        super().__init__()
        self.time = time
        self.step = timedelta(days=1)
        self.info = info

        self.out_info = None

    def _next_time(self):
        return self.time + self.step

    def _initialize(self):
        self.outputs.add(name="Output", info=self.info)
        self.create_connector()

    def _connect(self, start_time):
        self.try_connect(start_time, push_data={"Output": 1})
        self.out_info = self.connector.out_infos["Output"]

    def _validate(self):
        pass

    def _update(self):
        self.time += self.step
        self.outputs["Output"].push_data(1, self.time)

    def _finalize(self):
        pass


class SpecAdapter(Adapter):
    def __init__(self):
        super().__init__()

    def _get_data(self, time, target):
        return self.pull_data(time, target)

    def _get_info(self, info):
        in_info = self.exchange_info(info)
        out_info = in_info.copy_with(grid=info.grid)
        return out_info


class TestPropagate(unittest.TestCase):
    def test_propagate_info(self):
        time = datetime(2000, 1, 1)
        source = CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: 1,
                    Info(time=time, grid=NoGrid(), meta={"units": "m"}),
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = MockupConsumer(
            time, Info(time=time, grid=NoGrid(), meta={"units": "km"})
        )

        composition = Composition([source, sink])

        (source.outputs["Output"] >> SpecAdapter() >> sink.inputs["Input"])

        composition.run(end_time=datetime(2000, 1, 2))

        self.assertEqual(
            sink.inputs["Input"].info,
            Info(time=time, grid=NoGrid(), meta={"units": "km"}),
        )

    def test_propagate_info_fail(self):
        time = datetime(2000, 1, 1)
        source = CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: 1,
                    Info(time=time, grid=NoGrid(), meta={"units": "m"}),
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = MockupConsumer(
            time, Info(time=time, grid=NoGrid(), meta={"units": "m**2"})
        )

        composition = Composition([source, sink])

        source.outputs["Output"] >> sink.inputs["Input"]

        with self.assertRaises(FinamMetaDataError):
            composition.run(start_time=time, end_time=datetime(2000, 1, 2))

    def test_propagate_info_from_source(self):
        time = datetime(2000, 1, 1)
        source = CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: 1,
                    Info(time=time, grid=NoGrid(), meta={"units": "m"}),
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = MockupConsumer(
            datetime(2000, 1, 1), Info(time=None, grid=None, meta={"units": None})
        )

        composition = Composition([source, sink])

        source.outputs["Output"] >> sink.inputs["Input"]

        composition.run(start_time=time, end_time=datetime(2000, 1, 2))

        self.assertEqual(
            sink.inputs["Input"].info,
            Info(time=time, grid=NoGrid(), meta={"units": "m"}),
        )
        self.assertEqual(
            sink.inputs["Input"].info.time,
            time,
        )

    def test_propagate_info_from_target(self):
        time = datetime(2000, 1, 1)
        source = MockupProducer(
            time=datetime(2000, 1, 1),
            info=Info(time=None, grid=None, meta={"units": None}),
        )

        sink = MockupConsumer(
            datetime(2000, 1, 1), Info(time=time, grid=NoGrid(), meta={"units": "m"})
        )

        composition = Composition([source, sink])

        source.outputs["Output"] >> sink.inputs["Input"]

        composition.run(start_time=time, end_time=datetime(2000, 1, 2))

        self.assertEqual(
            source.outputs["Output"].info,
            Info(time=time, grid=NoGrid(), meta={"units": "m"}),
        )
        self.assertEqual(
            source.outputs["Output"].info.time,
            time,
        )
        self.assertEqual(
            source.out_info,
            Info(time=time, grid=NoGrid(), meta={"units": "m"}),
        )
        self.assertEqual(
            source.out_info.time,
            time,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import finam as fm


def u(s):
    return str(fm.UNITS.Unit(s))


class TestCfUnits(unittest.TestCase):
    def test_cf_units(self):
        self.assertEqual(u("mm"), "mm")
        self.assertEqual(u("millimeter"), "mm")
        self.assertEqual(u("kilometer"), "km")

        self.assertEqual(u("meter ** 2"), "m2")
        self.assertEqual(u("1 / meter"), "m-1")

        self.assertEqual(u("degree_north"), "degrees_north")
        self.assertEqual(u("degree_east"), "degrees_east")

        self.assertEqual(u("°C"), "°C")
        self.assertEqual(u("degC"), "°C")
        self.assertEqual(u("degree_Celsius"), "°C")

        # "dimensionless" representation inconsistent across different pint versions
        # will be "1" for newer versions (>=0.24.1) to be in line with cf-conventions
        self.assertTrue(u("") in ("1", "dimensionless", ""))
        self.assertTrue(u("1") in ("1", "dimensionless", ""))
        self.assertTrue(u("m/m") in ("1", "dimensionless", ""))

        self.assertEqual(u("m/s"), "m s-1")
        self.assertEqual(u("m s-1"), "m s-1")


if __name__ == "__main__":
    unittest.main()
